import os
import sys
import ctypes
import pytest

import sdl2
from sdl2 import SDL_Init, SDL_Quit, version, surface, rwops, render

sdlimage = pytest.importorskip("sdl2.sdlimage")


is32bit = sys.maxsize <= 2**32
ismacos = sys.platform == "darwin"
iswindows = "win" in sys.platform
isconda = os.getenv("CONDA_PREFIX") != None

formats = ["bmp",
           "cur",
           "gif",
           "ico",
           "jpg",
           "lbm",
           "pbm",
           "pcx",
           "pgm",
           "png",
           "pnm",
           "ppm",
           "svg",
           "tga",
           "tif",
           "webp",
           "xcf",
           "xpm",
           #"xv",
           ]

# SVG unsupported on SDL2_image < 2.0.2 as well as in Conda's current (2.0.5)
# Windows binaries
if sdlimage.dll.version < 2002 or (isconda and iswindows):
    formats.remove("svg")

# As of SDL2_image 2.0.5, XCF support seems to be broken on 32-bit builds
# XCF support is also broken in official SDL2_image macOS .frameworks and
# Conda's SDL2_image Windows binaries
bad_xcf = False
if is32bit or ismacos or (isconda and iswindows):
    formats.remove("xcf")
    bad_xcf = True

# WEBP support seems to be broken in the 32-bit Windows SDL2_image 2.0.2 binary
bad_webp = is32bit and sdlimage.dll.version == 2002
if bad_webp:
    formats.remove("webp")

# JPG saving requires SDL2_image >= 2.0.2 and isn't in official mac binaries
no_jpeg_save = sdlimage.dll.version < 2002 or ismacos

@pytest.fixture(scope="module")
def with_sdl_image(with_sdl):
    flags = (
        sdlimage.IMG_INIT_JPG | sdlimage.IMG_INIT_PNG |
        sdlimage.IMG_INIT_TIF  | sdlimage.IMG_INIT_WEBP
    )
    sdlimage.IMG_Init(flags)
    yield
    sdlimage.IMG_Quit()

@pytest.fixture
def sw_renderer(with_sdl_image):
    sf = surface.SDL_CreateRGBSurface(0, 10, 10, 32, 0, 0, 0, 0)
    rd = render.SDL_CreateSoftwareRenderer(sf)
    yield rd
    render.SDL_DestroyRenderer(rd)
    surface.SDL_FreeSurface(sf)

def _get_image_path(fmt):
    fname = "surfacetest.{0}".format(fmt)
    testdir = os.path.dirname(os.path.abspath(__file__))
    return os.path.join(testdir, "resources", fname)

def _verify_img_load(surf):
    if not surf:
        assert sdlimage.IMG_GetError() == b""
        assert surf
    assert isinstance(surf.contents, surface.SDL_Surface)


def test_IMG_Linked_Version():
    v = sdlimage.IMG_Linked_Version()
    assert isinstance(v.contents, version.SDL_version)
    assert v.contents.major == 2
    assert v.contents.minor == 0
    assert v.contents.patch >= 1

def test_IMG_Init():
    SDL_Init(0)
    supported = []
    libs = {
        'JPEG': sdlimage.IMG_INIT_JPG,
        'PNG': sdlimage.IMG_INIT_PNG,
        'TIFF': sdlimage.IMG_INIT_TIF,
        'WEBP': sdlimage.IMG_INIT_WEBP
    }
    for lib in libs.keys():
        flags = libs[lib]
        ret = sdlimage.IMG_Init(flags)
        err = sdlimage.IMG_GetError()
        if err:
            err = err.decode('utf-8')
            print("Error loading {0} library: {1}".format(lib, err))
            sdl2.SDL_ClearError()
        if ret & flags == flags:
            supported.append(lib)
        sdlimage.IMG_Quit()
    print("Supported image libraries:")
    print(supported)
    assert len(supported) == len(libs.keys())

def test_IMG_Load(with_sdl_image):
    for fmt in formats:
        fpath = _get_image_path(fmt)
        sf = sdlimage.IMG_Load(fpath.encode("utf-8"))
        _verify_img_load(sf)
        surface.SDL_FreeSurface(sf)

def test_IMG_Load_RW(with_sdl_image):
    skip = ['tga'] # TGA broken for Load_RW
    for fmt in formats:
        if fmt in skip:
            continue
        fpath = _get_image_path(fmt)
        with open(fpath, "rb") as fp:
            sf = sdlimage.IMG_Load_RW(rwops.rw_from_object(fp), False)
        _verify_img_load(sf)
        surface.SDL_FreeSurface(sf)

def test_IMG_LoadTexture(sw_renderer):
    rd = sw_renderer
    skip = []
    for fmt in formats:
        if fmt in skip:
            continue
        fpath = _get_image_path(fmt)
        tex = sdlimage.IMG_LoadTexture(rd, fpath.encode("utf-8"))
        if not tex:
            assert sdlimage.IMG_GetError() == b""
            assert tex
        assert isinstance(tex.contents, render.SDL_Texture)
        render.SDL_DestroyTexture(tex)

def test_IMG_LoadTexture_RW(sw_renderer):
    rd = sw_renderer
    skip = ['svg', 'tga'] # TGA & SVG broken for LoadTexture_RW
    for fmt in formats:
        if fmt in skip:
            continue
        fpath = _get_image_path(fmt)
        with open(fpath, "rb") as fp:
            tex = sdlimage.IMG_LoadTexture_RW(rd, rwops.rw_from_object(fp), 0)
            if not tex:
                assert sdlimage.IMG_GetError() == b""
                assert tex
            assert isinstance(tex.contents, render.SDL_Texture)
            render.SDL_DestroyTexture(tex)

def test_IMG_LoadTextureTyped_RW(sw_renderer):
    rd = sw_renderer
    skip = ['svg'] # SVG broken for LoadTextureTyped_RW
    for fmt in formats:
        if fmt in skip:
            continue
        fpath = _get_image_path(fmt)
        with open(fpath, "rb") as fp:
            rw = rwops.rw_from_object(fp)
            fmtx = fmt.upper().encode("utf-8")
            tex = sdlimage.IMG_LoadTextureTyped_RW(rd, rw, 0, fmtx)
            if not tex:
                assert sdlimage.IMG_GetError() == b""
                assert tex
            assert isinstance(tex.contents, render.SDL_Texture)
        render.SDL_DestroyTexture(tex)

def test_IMG_LoadTyped_RW(with_sdl_image):
    skip = []
    for fmt in formats:
        if fmt in skip:
            continue
        fpath = _get_image_path(fmt)
        with open(fpath, "rb") as fp:
            sf = sdlimage.IMG_LoadTyped_RW(
                rwops.rw_from_object(fp), False, fmt.upper().encode("utf-8")
            )
            _verify_img_load(sf)
            surface.SDL_FreeSurface(sf)

def test_IMG_LoadBMP_RW(with_sdl_image):
    fp = open(_get_image_path("bmp"), "rb")
    sf = sdlimage.IMG_LoadBMP_RW(rwops.rw_from_object(fp))
    fp.close()
    _verify_img_load(sf)
    surface.SDL_FreeSurface(sf)

def test_IMG_LoadCUR_RW(with_sdl_image):
    fp = open(_get_image_path("cur"), "rb")
    sf = sdlimage.IMG_LoadCUR_RW(rwops.rw_from_object(fp))
    fp.close()
    _verify_img_load(sf)
    surface.SDL_FreeSurface(sf)

def test_IMG_LoadGIF_RW(with_sdl_image):
    fp = open(_get_image_path("gif"), "rb")
    sf = sdlimage.IMG_LoadGIF_RW(rwops.rw_from_object(fp))
    fp.close()
    _verify_img_load(sf)
    surface.SDL_FreeSurface(sf)

def test_IMG_LoadICO_RW(with_sdl_image):
    fp = open(_get_image_path("ico"), "rb")
    sf = sdlimage.IMG_LoadICO_RW(rwops.rw_from_object(fp))
    fp.close()
    _verify_img_load(sf)
    surface.SDL_FreeSurface(sf)

def test_IMG_LoadJPG_RW(with_sdl_image):
    fp = open(_get_image_path("jpg"), "rb")
    sf = sdlimage.IMG_LoadJPG_RW(rwops.rw_from_object(fp))
    fp.close()
    _verify_img_load(sf)
    surface.SDL_FreeSurface(sf)

def test_IMG_LoadLBM_RW(with_sdl_image):
    fp = open(_get_image_path("lbm"), "rb")
    sf = sdlimage.IMG_LoadLBM_RW(rwops.rw_from_object(fp))
    fp.close()
    _verify_img_load(sf)
    surface.SDL_FreeSurface(sf)

def test_IMG_LoadPCX_RW(with_sdl_image):
    fp = open(_get_image_path("pcx"), "rb")
    sf = sdlimage.IMG_LoadPCX_RW(rwops.rw_from_object(fp))
    fp.close()
    _verify_img_load(sf)
    surface.SDL_FreeSurface(sf)

def test_IMG_LoadPNG_RW(with_sdl_image):
    fp = open(_get_image_path("png"), "rb")
    sf = sdlimage.IMG_LoadPNG_RW(rwops.rw_from_object(fp))
    fp.close()
    _verify_img_load(sf)
    surface.SDL_FreeSurface(sf)

def test_IMG_LoadPNM_RW(with_sdl_image):
    fp = open(_get_image_path("pnm"), "rb")
    sf = sdlimage.IMG_LoadPNM_RW(rwops.rw_from_object(fp))
    fp.close()
    _verify_img_load(sf)
    surface.SDL_FreeSurface(sf)

@pytest.mark.skipif(sdlimage.dll.version < 2002, reason="Added in 2.0.2")
@pytest.mark.xfail(isconda and iswindows, reason="Broken w/ win64 Conda")
def test_IMG_LoadSVG_RW(with_sdl_image):
    fp = open(_get_image_path("svg"), "rb")
    sf = sdlimage.IMG_LoadSVG_RW(rwops.rw_from_object(fp))
    fp.close()
    _verify_img_load(sf)
    surface.SDL_FreeSurface(sf)

def test_IMG_LoadTGA_RW(with_sdl_image):
    fp = open(_get_image_path("tga"), "rb")
    sf = sdlimage.IMG_LoadTGA_RW(rwops.rw_from_object(fp))
    fp.close()
    _verify_img_load(sf)
    surface.SDL_FreeSurface(sf)

def test_IMG_LoadTIF_RW(with_sdl_image):
    fp = open(_get_image_path("tif"), "rb")
    sf = sdlimage.IMG_LoadTIF_RW(rwops.rw_from_object(fp))
    fp.close()
    _verify_img_load(sf)
    surface.SDL_FreeSurface(sf)

@pytest.mark.xfail(bad_webp, reason="WEBP broken in 2.0.2 binaries for 32-bit Windows")
def test_IMG_LoadWEBP_RW(with_sdl_image):
    fp = open(_get_image_path("webp"), "rb")
    sf = sdlimage.IMG_LoadWEBP_RW(rwops.rw_from_object(fp))
    fp.close()
    _verify_img_load(sf)
    surface.SDL_FreeSurface(sf)

@pytest.mark.xfail(bad_xcf, reason="XCF currently broken on some platforms")
def test_IMG_LoadXCF_RW(with_sdl_image):
    fp = open(_get_image_path("xcf"), "rb")
    sf = sdlimage.IMG_LoadXCF_RW(rwops.rw_from_object(fp))
    fp.close()
    _verify_img_load(sf)
    surface.SDL_FreeSurface(sf)

def test_IMG_LoadXPM_RW(with_sdl_image):
    fp = open(_get_image_path("xpm"), "rb")
    sf = sdlimage.IMG_LoadXPM_RW(rwops.rw_from_object(fp))
    fp.close()
    _verify_img_load(sf)
    surface.SDL_FreeSurface(sf)

@pytest.mark.skip("not implemented")
def test_IMG_LoadXV_RW(with_sdl_image):
    fp = open(_get_image_path("xv"), "rb")
    sf = sdlimage.IMG_LoadXV_RW(rwops.rw_from_object(fp))
    fp.close()
    _verify_img_load(sf)
    surface.SDL_FreeSurface(sf)

def test_IMG_isBMP(with_sdl_image):
    for fmt in formats:
        fpath = _get_image_path(fmt)
        with open(fpath, "rb") as fp:
            imgrw = rwops.rw_from_object(fp)
            if fmt == "bmp":
                assert sdlimage.IMG_isBMP(imgrw)
            else:
                assert not sdlimage.IMG_isBMP(imgrw)

def test_IMG_isCUR(with_sdl_image):
    for fmt in formats:
        fpath = _get_image_path(fmt)
        with open(fpath, "rb") as fp:
            imgrw = rwops.rw_from_object(fp)
            if fmt == "cur":
                assert sdlimage.IMG_isCUR(imgrw)
            else:
                assert not sdlimage.IMG_isCUR(imgrw)

def test_IMG_isGIF(with_sdl_image):
    for fmt in formats:
        fpath = _get_image_path(fmt)
        with open(fpath, "rb") as fp:
            imgrw = rwops.rw_from_object(fp)
            if fmt == "gif":
                assert sdlimage.IMG_isGIF(imgrw)
            else:
                assert not sdlimage.IMG_isGIF(imgrw)

def test_IMG_isICO(with_sdl_image):
    for fmt in formats:
        fpath = _get_image_path(fmt)
        with open(fpath, "rb") as fp:
            imgrw = rwops.rw_from_object(fp)
            if fmt == "ico":
                assert sdlimage.IMG_isICO(imgrw)
            else:
                assert not sdlimage.IMG_isICO(imgrw)

def test_IMG_isJPG(with_sdl_image):
    for fmt in formats:
        fpath = _get_image_path(fmt)
        with open(fpath, "rb") as fp:
            imgrw = rwops.rw_from_object(fp)
            if fmt == "jpg":
                assert sdlimage.IMG_isJPG(imgrw)
            else:
                assert not sdlimage.IMG_isJPG(imgrw)

def test_IMG_isLBM(with_sdl_image):
    for fmt in formats:
        fpath = _get_image_path(fmt)
        with open(fpath, "rb") as fp:
            imgrw = rwops.rw_from_object(fp)
            if fmt == "lbm":
                assert sdlimage.IMG_isLBM(imgrw)
            else:
                assert not sdlimage.IMG_isLBM(imgrw)

def test_IMG_isPCX(with_sdl_image):
    for fmt in formats:
        fpath = _get_image_path(fmt)
        with open(fpath, "rb") as fp:
            imgrw = rwops.rw_from_object(fp)
            if fmt == "pcx":
                assert sdlimage.IMG_isPCX(imgrw)
            else:
                assert not sdlimage.IMG_isPCX(imgrw)

def test_IMG_isPNG(with_sdl_image):
    for fmt in formats:
        fpath = _get_image_path(fmt)
        with open(fpath, "rb") as fp:
            imgrw = rwops.rw_from_object(fp)
            if fmt == "png":
                assert sdlimage.IMG_isPNG(imgrw)
            else:
                assert not sdlimage.IMG_isPNG(imgrw)

def test_IMG_isPNM(with_sdl_image):
    for fmt in formats:
        fpath = _get_image_path(fmt)
        with open(fpath, "rb") as fp:
            imgrw = rwops.rw_from_object(fp)
            if fmt in ("pnm", "pbm", "ppm", "pgm"):
                assert sdlimage.IMG_isPNM(imgrw)
            else:
                assert not sdlimage.IMG_isPNM(imgrw)

@pytest.mark.skipif(sdlimage.dll.version < 2002, reason="Added in 2.0.2")
@pytest.mark.xfail(isconda and iswindows, reason="Broken w/ win64 Conda")
def test_IMG_isSVG(with_sdl_image):
    for fmt in formats:
        fpath = _get_image_path(fmt)
        with open(fpath, "rb") as fp:
            imgrw = rwops.rw_from_object(fp)
            if fmt == "svg":
                assert sdlimage.IMG_isSVG(imgrw)
            else:
                assert not sdlimage.IMG_isSVG(imgrw)

def test_IMG_isTIF(with_sdl_image):
    for fmt in formats:
        fpath = _get_image_path(fmt)
        with open(fpath, "rb") as fp:
            imgrw = rwops.rw_from_object(fp)
            if fmt == "tif":
                assert sdlimage.IMG_isTIF(imgrw)
            else:
                assert not sdlimage.IMG_isTIF(imgrw)

def test_IMG_isWEBP(with_sdl_image):
    for fmt in formats:
        fpath = _get_image_path(fmt)
        with open(fpath, "rb") as fp:
            imgrw = rwops.rw_from_object(fp)
            if fmt == "webp":
                assert sdlimage.IMG_isWEBP(imgrw)
            else:
                assert not sdlimage.IMG_isWEBP(imgrw)

@pytest.mark.xfail(bad_xcf, reason="XCF currently broken on some platforms")
def test_IMG_isXCF(with_sdl_image):
    for fmt in formats:
        fpath = _get_image_path(fmt)
        with open(fpath, "rb") as fp:
            imgrw = rwops.rw_from_object(fp)
            if fmt == "xcf":
                assert sdlimage.IMG_isXCF(imgrw)
            else:
                assert not sdlimage.IMG_isXCF(imgrw)

def test_IMG_isXPM(with_sdl_image):
    for fmt in formats:
        fpath = _get_image_path(fmt)
        with open(fpath, "rb") as fp:
            imgrw = rwops.rw_from_object(fp)
            if fmt == "xpm":
                assert sdlimage.IMG_isXPM(imgrw)
            else:
                assert not sdlimage.IMG_isXPM(imgrw)

@pytest.mark.skip("not implemented")
def test_IMG_isXV(with_sdl_image):
    for fmt in formats:
        fpath = _get_image_path(fmt)
        with open(fpath, "rb") as fp:
            imgrw = rwops.rw_from_object(fp)
            if fmt == "xv":
                assert sdlimage.IMG_isXV(imgrw)
            else:
                assert not sdlimage.IMG_isXV(imgrw)

@pytest.mark.skipif(hasattr(sys, "pypy_version_info"),
    reason="PyPy's ctypes fails to pass a correct string array")
def test_IMG_ReadXPMFromArray(with_sdl_image):
    fp = open(_get_image_path("xpm"), "rb")
    xpm = b""
    fp.readline()  # /* XPM */
    fp.readline()  # static char * surfacetest_xpm[] = {
    lbuf = fp.readlines()
    fp.close()
    for line in lbuf:
        if line.endswith(b"};"):
            xpm += line[1:-4]
        else:
            xpm += line[1:-3]
    pxpm = ctypes.c_char_p(xpm)
    sf = sdlimage.IMG_ReadXPMFromArray(ctypes.byref(pxpm))
    _verify_img_load(sf)
    surface.SDL_FreeSurface(sf)

def test_IMG_SavePNG(tmpdir):
    # Open a PNG that we can re-save
    fpath = _get_image_path("png")
    sf = sdlimage.IMG_Load(fpath.encode("utf-8"))
    assert isinstance(sf.contents, surface.SDL_Surface)

    # Try saving the PNG to a new folder
    outpath = os.path.join(str(tmpdir), "save_test.png")
    ret = sdlimage.IMG_SavePNG(sf, outpath.encode("utf-8"))
    assert sdl2.SDL_GetError() == b""
    assert ret == 0
    assert os.path.exists(outpath)
    surface.SDL_FreeSurface(sf)

@pytest.mark.skip("not working yet")
def test_IMG_SavePNG_RW(tmpdir):
    # Open a PNG that we can re-save
    fpath = _get_image_path("png")
    sf = sdlimage.IMG_Load(fpath.encode("utf-8"))
    assert isinstance(sf.contents, surface.SDL_Surface)

    # Try saving the PNG to a new folder
    outpath = os.path.join(str(tmpdir), "save_test.png")
    rw = rwops.SDL_RWFromFile(outpath.encode("utf-8"), b"wb")
    ret = sdlimage.IMG_SavePNG_RW(sf, rw, 0)
    assert sdl2.SDL_GetError() == b""
    assert ret == 0
    assert os.path.exists(outpath)

    # Try reopening the RW as a PNG
    sf2 = sdlimage.IMG_LoadPNG_RW(rw)
    assert sdl2.SDL_GetError() == b""
    assert isinstance(sf2.contents, surface.SDL_Surface)
    surface.SDL_FreeSurface(sf)
    surface.SDL_FreeSurface(sf2)

@pytest.mark.skipif(no_jpeg_save, reason="Added in 2.0.2, not in macOS bnaries")
def test_IMG_SaveJPG(tmpdir):
    # Open a PNG that we can save to JPG
    fpath = _get_image_path("png")
    sf = sdlimage.IMG_Load(fpath.encode("utf-8"))
    assert isinstance(sf.contents, surface.SDL_Surface)

    # Try saving as JPG to a new folder
    outpath = os.path.join(str(tmpdir), "save_test.jpg")
    ret = sdlimage.IMG_SaveJPG(sf, outpath.encode("utf-8"), 90)
    assert sdl2.SDL_GetError() == b""
    assert ret == 0
    assert os.path.exists(outpath)
    surface.SDL_FreeSurface(sf)

@pytest.mark.skip("not working yet")
@pytest.mark.skipif(no_jpeg_save, reason="Added in 2.0.2, not in macOS bnaries")
def test_IMG_SaveJPG_RW(tmpdir):
    # Open a PNG that we can save to JPG
    fpath = _get_image_path("png")
    sf = sdlimage.IMG_Load(fpath.encode("utf-8"))
    assert isinstance(sf.contents, surface.SDL_Surface)

    # Try saving as JPG to a new folder
    outpath = os.path.join(str(tmpdir), "save_test.jpg")
    rw = rwops.SDL_RWFromFile(outpath.encode("utf-8"), b"wb")
    ret = sdlimage.IMG_SaveJPG_RW(sf, rw, 0, 90)
    assert sdl2.SDL_GetError() == b""
    assert ret == 0
    assert os.path.exists(outpath)

    # Try reopening the RW as a JPG
    sf2 = sdlimage.IMG_LoadJPG_RW(rw)
    assert sdl2.SDL_GetError() == b""
    assert isinstance(sf2.contents, surface.SDL_Surface)
    surface.SDL_FreeSurface(sf)
    surface.SDL_FreeSurface(sf2)



# -*- coding: utf-8 -*-
import os
import gc
import pytest
from struct import unpack
from ctypes import byref, c_int, c_uint16
import sdl2
from sdl2 import SDL_TRUE, SDL_FALSE, SDL_Color, surface, version, rwops
from sdl2.ext.compat import utf8

sdlttf = pytest.importorskip("sdl2.sdlttf")

parent_dir = os.path.dirname(os.path.abspath(__file__))
fontfile = os.path.join(parent_dir, "resources", "tuffy.ttf").encode("utf-8")
font_test_sizes = [6, 16, 26]

def to_utf16(x):
    # Converts a unicode Python string to a ctypes UTF-16 array
    strlen = len(x) + 1 # +1 for byte-order mark
    intstr = unpack('H' * strlen, utf8(x).encode('utf-16')) 
    intstr = intstr + (0, ) # Add null byte at end to terminate string
    return (c_uint16 * (strlen + 1))(*intstr)

@pytest.fixture(scope="module")
def with_sdl_ttf(with_sdl):
    ret = sdlttf.TTF_Init()
    assert sdlttf.TTF_GetError() == b""
    assert ret == 0
    yield
    while sdlttf.TTF_WasInit() > 0:
        sdlttf.TTF_Quit()

@pytest.fixture()
def with_font(with_sdl_ttf):
    sdl2.SDL_ClearError()
    font = sdlttf.TTF_OpenFont(fontfile, 20)
    assert sdlttf.TTF_GetError() == b""
    assert font
    yield font
    sdlttf.TTF_CloseFont(font)


def test_TTF_Font():
    font = sdlttf.TTF_Font()
    assert isinstance(font, sdlttf.TTF_Font)

def test_TTF_InitQuit(with_sdl):
    # Test that init and quit actually work
    ret = sdlttf.TTF_Init()
    assert sdlttf.TTF_GetError() == b""
    assert ret == 0
    assert sdlttf.TTF_WasInit()
    sdlttf.TTF_Quit()
    assert not sdlttf.TTF_WasInit()
    # Every time TTF_Init() is run, internal number increments by 1,
    # every time TTF_Quit() is run, internal number decrements by 1 and
    # only actually quits when internal number == 0
    sdlttf.TTF_Init()
    sdlttf.TTF_Init()
    sdlttf.TTF_Quit()
    assert sdlttf.TTF_WasInit()
    sdlttf.TTF_Quit()
    assert not sdlttf.TTF_WasInit()

def test_TTF_Linked_Version(with_sdl_ttf):
    v = sdlttf.TTF_Linked_Version()
    assert isinstance(v.contents, version.SDL_version)
    assert v.contents.major == 2
    assert v.contents.minor == 0
    assert v.contents.patch >= 12

@pytest.mark.skipif(sdlttf.dll.version < 2018, reason="not available")
def test_TTF_GetFreeTypeVersion(with_sdl_ttf):
    major, minor, patch = c_int(0), c_int(0), c_int(0)
    sdlttf.TTF_GetFreeTypeVersion(byref(major), byref(minor), byref(patch))
    assert major.value > 0  # Only one guaranteed to be non-zero

@pytest.mark.skipif(sdlttf.dll.version < 2018, reason="not available")
def test_TTF_GetHarfBuzzVersion(with_sdl_ttf):
    major, minor, patch = c_int(0), c_int(0), c_int(0)
    sdlttf.TTF_GetHarfBuzzVersion(byref(major), byref(minor), byref(patch))
    assert major.value > 0  # Only one guaranteed to be non-zero

def test_TTF_ByteSwappedUNICODE(with_sdl_ttf):
    sdlttf.TTF_ByteSwappedUNICODE(0)
    sdlttf.TTF_ByteSwappedUNICODE(1)

def test_TTF_OpenCloseFont(with_sdl_ttf):
    for x in font_test_sizes:
        font = sdlttf.TTF_OpenFont(fontfile, x)
        assert sdlttf.TTF_GetError() == b""
        assert isinstance(font.contents, sdlttf.TTF_Font)
        sdlttf.TTF_CloseFont(font)

def test_TTF_OpenFontIndex(with_sdl_ttf):
    for x in font_test_sizes:
        font = sdlttf.TTF_OpenFontIndex(fontfile, x, 0)
        assert sdlttf.TTF_GetError() == b""
        assert isinstance(font.contents, sdlttf.TTF_Font)
        sdlttf.TTF_CloseFont(font)

def test_TTF_OpenFontRW(with_sdl_ttf):
    fp = open(fontfile, "rb")
    fontrw = rwops.rw_from_object(fp)
    for x in font_test_sizes:
        fp.seek(0)
        font = sdlttf.TTF_OpenFontRW(fontrw, 0, x)
        assert sdlttf.TTF_GetError() == b""
        assert isinstance(font.contents, sdlttf.TTF_Font)
        sdlttf.TTF_CloseFont(font)
    fp.close()

def test_TTF_OpenFontIndexRW(with_sdl_ttf):
    fp = open(fontfile, "rb")
    fontrw = rwops.rw_from_object(fp)
    for x in font_test_sizes:
        fp.seek(0)
        font = sdlttf.TTF_OpenFontIndexRW(fontrw, 0, x, 0)
        assert sdlttf.TTF_GetError() == b""
        assert isinstance(font.contents, sdlttf.TTF_Font)
        sdlttf.TTF_CloseFont(font)
    fp.close()

@pytest.mark.skipif(sdlttf.dll.version < 2018, reason="not available")
def test_TTF_OpenFontDPI(with_sdl_ttf):
    # This actually tests for DPI having an effect on the text, the other
    # OpenFontDPI tests just test simple loading
    w1, h1, w2, h2 = c_int(0), c_int(0), c_int(0), c_int(0)
    font = sdlttf.TTF_OpenFontDPI(fontfile, 30, 80, 80)
    assert sdlttf.TTF_GetError() == b""
    assert isinstance(font.contents, sdlttf.TTF_Font)
    sdlttf.TTF_SizeText(font, b"Hi there!", byref(w1), byref(h1))
    sdlttf.TTF_CloseFont(font)
    font = sdlttf.TTF_OpenFontDPI(fontfile, 30, 100, 60)
    assert sdlttf.TTF_GetError() == b""
    sdlttf.TTF_SizeText(font, b"Hi there!", byref(w2), byref(h2))
    sdlttf.TTF_CloseFont(font)
    # Make sure text size differs between DPIs
    assert w2.value > w1.value
    assert h2.value < h1.value

@pytest.mark.skipif(sdlttf.dll.version < 2018, reason="not available")
def test_TTF_OpenFontIndexDPI(with_sdl_ttf):
    test_dpi_sizes = [(50, 50), (80, 40), (100, 100)]
    for hdpi, vdpi in test_dpi_sizes:
        font = sdlttf.TTF_OpenFontIndexDPI(fontfile, 30, 0, hdpi, vdpi)
        assert sdlttf.TTF_GetError() == b""
        assert isinstance(font.contents, sdlttf.TTF_Font)
        sdlttf.TTF_CloseFont(font)

@pytest.mark.skipif(sdlttf.dll.version < 2018, reason="not available")
def test_TTF_OpenFontDPIRW(with_sdl_ttf):
    test_dpi_sizes = [(50, 50), (80, 40), (100, 100)]
    fp = open(fontfile, "rb")
    fontrw = rwops.rw_from_object(fp)
    for hdpi, vdpi in test_dpi_sizes:
        font = sdlttf.TTF_OpenFontDPIRW(fontrw, 0, 30, hdpi, vdpi)
        assert sdlttf.TTF_GetError() == b""
        assert isinstance(font.contents, sdlttf.TTF_Font)
        sdlttf.TTF_CloseFont(font)    

@pytest.mark.skipif(sdlttf.dll.version < 2018, reason="not available")
def test_TTF_OpenFontIndexDPIRW(with_sdl_ttf):
    test_dpi_sizes = [(50, 50), (80, 40), (100, 100)]
    fp = open(fontfile, "rb")
    fontrw = rwops.rw_from_object(fp)
    for hdpi, vdpi in test_dpi_sizes:
        font = sdlttf.TTF_OpenFontIndexDPIRW(fontrw, 0, 30, 0, hdpi, vdpi)
        assert sdlttf.TTF_GetError() == b""
        assert isinstance(font.contents, sdlttf.TTF_Font)
        sdlttf.TTF_CloseFont(font)  

@pytest.mark.skipif(sdlttf.dll.version < 2018, reason="not available")
def test_TTF_SetFontSize(with_font):
    font = with_font
    w1, h1, w2, h2 = c_int(0), c_int(0), c_int(0), c_int(0)
    sdlttf.TTF_SizeText(font, b"Hi there!", byref(w1), byref(h1))
    # Increase font size and make sure it works (original size is 20pt)
    ret = sdlttf.TTF_SetFontSize(font, 30)
    assert sdlttf.TTF_GetError() == b""
    assert ret == 0
    sdlttf.TTF_SizeText(font, b"Hi there!", byref(w2), byref(h2))
    assert w2.value > w1.value
    assert h2.value > h1.value
    # Decrease font size and make sure it works
    ret = sdlttf.TTF_SetFontSize(font, 10)
    assert sdlttf.TTF_GetError() == b""
    assert ret == 0
    sdlttf.TTF_SizeText(font, b"Hi there!", byref(w2), byref(h2))
    assert w2.value < w1.value
    assert h2.value < h1.value

@pytest.mark.skipif(sdlttf.dll.version < 2018, reason="not available")
def test_TTF_SetFontSizeDPI(with_font):
    font = with_font
    w1, h1, w2, h2 = c_int(0), c_int(0), c_int(0), c_int(0)
    sdlttf.TTF_SizeText(font, b"Hi there!", byref(w1), byref(h1))
    # Adjust and make sure it works (original DPI is 72)
    ret = sdlttf.TTF_SetFontSizeDPI(font, 20, 100, 50)
    assert sdlttf.TTF_GetError() == b""
    assert ret == 0
    sdlttf.TTF_SizeText(font, b"Hi there!", byref(w2), byref(h2))
    assert w2.value > w1.value
    assert h2.value < h1.value
    # Resize the text with a different DPI and see if it works
    ret = sdlttf.TTF_SetFontSizeDPI(font, 20, 50, 100)
    assert sdlttf.TTF_GetError() == b""
    assert ret == 0
    sdlttf.TTF_SizeText(font, b"Hi there!", byref(w2), byref(h2))
    assert w2.value < w1.value
    assert h2.value > h1.value

def test_TTF_GetSetFontStyle(with_font):
    normal = sdlttf.TTF_STYLE_NORMAL
    bold = sdlttf.TTF_STYLE_BOLD
    italic = sdlttf.TTF_STYLE_ITALIC
    underline = sdlttf.TTF_STYLE_UNDERLINE
    # Test out getting/setting different font styles
    font = with_font
    assert sdlttf.TTF_GetFontStyle(font) == normal
    sdlttf.TTF_SetFontStyle(font, bold)
    assert sdlttf.TTF_GetFontStyle(font) == bold
    sdlttf.TTF_SetFontStyle(font, bold | italic)
    assert sdlttf.TTF_GetFontStyle(font) == bold | italic
    sdlttf.TTF_SetFontStyle(font, bold | underline)
    assert sdlttf.TTF_GetFontStyle(font) == bold | underline

def test_TTF_GetSetFontOutline(with_font):
    font = with_font
    assert sdlttf.TTF_GetFontOutline(font) == 0
    for x in range(1, 11, 2):
        sdlttf.TTF_SetFontOutline(font, x)
        assert sdlttf.TTF_GetFontOutline(font) == x

def test_TTF_GetSetFontHinting(with_font):
    font = with_font
    hints = [
        sdlttf.TTF_HINTING_NORMAL, sdlttf.TTF_HINTING_LIGHT,
        sdlttf.TTF_HINTING_MONO, sdlttf.TTF_HINTING_NONE
    ]
    if sdlttf.dll.version >= 2018:
        hints.append(sdlttf.TTF_HINTING_LIGHT_SUBPIXEL)
    assert sdlttf.TTF_GetFontHinting(font) == sdlttf.TTF_HINTING_NORMAL
    for hint in hints:
        sdlttf.TTF_SetFontHinting(font, hint)
        assert sdlttf.TTF_GetFontHinting(font) == hint

def test_TTF_FontHeight(with_sdl_ttf):
    last = cur = 0
    for ptsize in font_test_sizes:
        font = sdlttf.TTF_OpenFont(fontfile, ptsize)
        cur = sdlttf.TTF_FontHeight(font)
        assert cur >= last
        last = cur
        sdlttf.TTF_CloseFont(font)

def test_TTF_FontAscent(with_sdl_ttf):
    last = cur = 0
    for ptsize in font_test_sizes:
        font = sdlttf.TTF_OpenFont(fontfile, ptsize)
        cur = sdlttf.TTF_FontAscent(font)
        assert cur >= last
        last = cur
        sdlttf.TTF_CloseFont(font)

def test_TTF_FontDescent(with_sdl_ttf):
    last = cur = 0
    for ptsize in font_test_sizes:
        font = sdlttf.TTF_OpenFont(fontfile, ptsize)
        cur = sdlttf.TTF_FontDescent(font)
        assert cur <= last
        last = cur
        sdlttf.TTF_CloseFont(font)

def test_TTF_FontLineSkip(with_sdl_ttf):
    last = cur = 0
    for ptsize in font_test_sizes:
        font = sdlttf.TTF_OpenFont(fontfile, ptsize)
        cur = sdlttf.TTF_FontLineSkip(font)
        assert cur >= last
        last = cur
        sdlttf.TTF_CloseFont(font)

def test_TTF_GetSetFontKerning(with_font):
    font = with_font
    assert sdlttf.TTF_GetFontKerning(font) == 1
    sdlttf.TTF_SetFontKerning(font, 0)
    assert sdlttf.TTF_GetFontKerning(font) == 0
    sdlttf.TTF_SetFontKerning(font, 1)
    assert sdlttf.TTF_GetFontKerning(font) == 1
    sdlttf.TTF_SetFontKerning(font, 0)
    assert sdlttf.TTF_GetFontKerning(font) == 0

def test_TTF_FontFaces(with_font):
    font = with_font
    assert sdlttf.TTF_FontFaces(font) >= 1

def test_TTF_FontFaceIsFixedWidth(with_font):
    font = with_font
    assert not sdlttf.TTF_FontFaceIsFixedWidth(font)

def test_TTF_FontFaceFamilyName(with_font):
    font = with_font
    assert sdlttf.TTF_FontFaceFamilyName(font) == b"Tuffy"

def test_TTF_FontFaceStyleName(with_font):
    font = with_font
    assert sdlttf.TTF_FontFaceStyleName(font) == b"Regular"

def test_TTF_GlyphIsProvided(with_font):
    font = with_font
    assert isinstance(font.contents, sdlttf.TTF_Font)
    for ch in range(32, 127):
        assert sdlttf.TTF_GlyphIsProvided(font, ch)
    assert not sdlttf.TTF_GlyphIsProvided(font, 0)
    assert not sdlttf.TTF_GlyphIsProvided(font, 0x0ff9)

@pytest.mark.skipif(sdlttf.dll.version < 2018, reason="not available")
def test_TTF_GlyphIsProvided32(with_font):
    font = with_font
    assert isinstance(font.contents, sdlttf.TTF_Font)
    for ch in range(32, 127):
        assert sdlttf.TTF_GlyphIsProvided32(font, ch)
    assert not sdlttf.TTF_GlyphIsProvided32(font, 0)
    assert not sdlttf.TTF_GlyphIsProvided32(font, 0x0ff9)

def test_TTF_GlyphMetrics(with_sdl_ttf):
    expected = {
        'A': [1, 25, 0, 29, 25],
        'j': [-3, 7, -9, 28, 9],
        '.': [2, 7, -1, 4, 8]
    }
    font = sdlttf.TTF_OpenFont(fontfile, 40)
    minX, maxX, minY, maxY = c_int(0), c_int(0), c_int(0), c_int(0)
    adv = c_int(0)
    for char in expected.keys():
        ret = sdlttf.TTF_GlyphMetrics(
            font, ord(char),
            byref(minX), byref(maxX), byref(minY), byref(maxY), byref(adv)
        )
        results = [x.value for x in (minX, maxX, minY, maxY, adv)]
        assert sdlttf.TTF_GetError() == b""
        assert ret == 0
        assert results == expected[char]
    sdlttf.TTF_CloseFont(font)

@pytest.mark.skipif(sdlttf.dll.version < 2018, reason="not available")
def test_TTF_GlyphMetrics32(with_sdl_ttf):
    expected = {
        'A': [1, 25, 0, 29, 25],
        'j': [-3, 7, -9, 28, 9],
        '.': [2, 7, -1, 4, 8]
    }
    font = sdlttf.TTF_OpenFont(fontfile, 40)
    minX, maxX, minY, maxY = c_int(0), c_int(0), c_int(0), c_int(0)
    adv = c_int(0)
    for char in expected.keys():
        ret = sdlttf.TTF_GlyphMetrics32(
            font, ord(char),
            byref(minX), byref(maxX), byref(minY), byref(maxY), byref(adv)
        )
        results = [x.value for x in (minX, maxX, minY, maxY, adv)]
        assert sdlttf.TTF_GetError() == b""
        assert ret == 0
        assert results == expected[char]
    sdlttf.TTF_CloseFont(font)

def test_TTF_SizeText(with_font):
    font = with_font
    min_expected_w = 69     # SDL2_ttf 2.0.18
    max_expected_w = 70     # SDL2_ttf <= 2.0.15
    min_expected_h = 21     # SDL2_ttf 2.0.15 with FreeType 2.10.1
    max_expected_h = 25     # SDL2_ttf < 2.0.15
    w, h = c_int(0), c_int(0)
    sdlttf.TTF_SizeText(font, b"Hi there!", byref(w), byref(h))
    assert w.value >= min_expected_w
    assert w.value <= max_expected_w
    assert h.value >= min_expected_h
    assert h.value <= max_expected_h

def test_TTF_SizeUTF8(with_font):
    font = with_font
    min_expected_w = 72     # SDL2_ttf 2.0.18
    max_expected_w = 73     # SDL2_ttf <= 2.0.15
    min_expected_h = 21     # SDL2_ttf 2.0.15 with FreeType 2.10.1
    max_expected_h = 25     # SDL2_ttf < 2.0.15
    w, h = c_int(0), c_int(0)
    sdlttf.TTF_SizeUTF8(font, u"Hï thère!".encode('utf-8'), byref(w), byref(h))
    assert w.value >= min_expected_w
    assert w.value <= max_expected_w
    assert h.value >= min_expected_h
    assert h.value <= max_expected_h

def test_TTF_SizeUNICODE(with_font):
    font = with_font
    min_expected_w = 69     # SDL2_ttf 2.0.18
    max_expected_w = 70     # SDL2_ttf <= 2.0.15
    min_expected_h = 21     # SDL2_ttf 2.0.15 with FreeType 2.10.1
    max_expected_h = 25     # SDL2_ttf < 2.0.15
    w, h = c_int(0), c_int(0)
    strarr = to_utf16(u"Hi there!")
    sdlttf.TTF_SizeUNICODE(font, strarr, byref(w), byref(h))
    # For debug purposes
    #print(list(strarr))
    #print("w = {0}, h = {1}".format(w.value, h.value))
    assert w.value >= min_expected_w
    assert w.value <= max_expected_w
    assert h.value >= min_expected_h
    assert h.value <= max_expected_h

@pytest.mark.skipif(sdlttf.dll.version < 2018, reason="not available")
def test_TTF_Measure(with_font):
    font = with_font
    extent, count = c_int(0), c_int(0)
    w, h = c_int(0), c_int(0)
    tst = b"This is a long line that should be wrapped!"
    # With TTF_MeasureText
    ret = sdlttf.TTF_MeasureText(font, tst, 200, byref(extent), byref(count))
    assert sdlttf.TTF_GetError() == b""
    assert ret == 0
    assert all([x.value > 0 for x in (extent, count)])
    sdlttf.TTF_SizeText(font, tst[:count.value], byref(w), byref(h))
    assert extent.value == w.value
    # With TTF_MeasureUTF8
    ret = sdlttf.TTF_MeasureUTF8(font, tst, 180, byref(extent), byref(count))
    assert sdlttf.TTF_GetError() == b""
    assert ret == 0
    assert all([x.value > 0 for x in (extent, count)])
    sdlttf.TTF_SizeUTF8(font, tst[:count.value], byref(w), byref(h))
    assert extent.value == w.value
    # With TTF_MeasureUNICODE
    strarr = to_utf16(tst)
    ret = sdlttf.TTF_MeasureUNICODE(font, strarr, 220, byref(extent), byref(count))
    assert sdlttf.TTF_GetError() == b""
    assert ret == 0
    assert all([x.value > 0 for x in (extent, count)])
    strarr = to_utf16(tst[:count.value])
    sdlttf.TTF_SizeUNICODE(font, strarr, byref(w), byref(h))
    assert extent.value == w.value

def test_TTF_Render_Solid(with_font):
    font = with_font
    color = SDL_Color(0, 0, 0)
    # Test TTF_RenderText_Solid
    sf = sdlttf.TTF_RenderText_Solid(font, b"Hi there!", color)
    assert isinstance(sf.contents, surface.SDL_Surface)
    # Test TTF_RenderUTF8_Solid
    teststr = u"Hï thère!".encode('utf-8')
    sf = sdlttf.TTF_RenderUTF8_Solid(font, teststr, color)
    assert isinstance(sf.contents, surface.SDL_Surface)
    # Test TTF_RenderUNICODE_Solid
    # NOTE: no unicode chars because number -> glyph lookup is os-dependent
    strarr = to_utf16(u"Hi there!")
    sf = sdlttf.TTF_RenderUNICODE_Solid(font, strarr, color)
    assert isinstance(sf.contents, surface.SDL_Surface)
    # Test TTF_RenderGlyph_Solid
    sf = sdlttf.TTF_RenderGlyph_Solid(font, ord("A"), color)
    assert isinstance(sf.contents, surface.SDL_Surface)

@pytest.mark.skipif(sdlttf.dll.version < 2018, reason="not available")
def test_TTF_Render_Solid_Wrapped(with_font):
    font = with_font
    color = SDL_Color(0, 0, 0, 255)
    # Test TTF_RenderText_Solid_Wrapped
    teststr = b"Hi there, this is a long line!"
    sf = sdlttf.TTF_RenderText_Solid_Wrapped(font, teststr, color, 100)
    assert isinstance(sf.contents, surface.SDL_Surface)
    assert sf.contents.h > 30
    # Test TTF_RenderUTF8_Solid_Wrapped
    teststr = u"Hï thère, this is a long line!".encode('utf-8')
    sf = sdlttf.TTF_RenderUTF8_Solid_Wrapped(font, teststr, color, 100)
    assert isinstance(sf.contents, surface.SDL_Surface)
    assert sf.contents.h > 30
    # Test TTF_RenderUNICODE_Solid_Wrapped
    # NOTE: no unicode chars because number -> glyph lookup is os-dependent
    strarr = to_utf16(u"Hi there, this is a long line!")
    sf = sdlttf.TTF_RenderUNICODE_Solid_Wrapped(font, strarr, color, 100)
    assert isinstance(sf.contents, surface.SDL_Surface)
    assert sf.contents.h > 30

def test_TTF_Render_Shaded(with_font):
    font = with_font
    color = SDL_Color(0, 0, 0)
    bgcolor = SDL_Color(255, 255, 255)
    # Test TTF_RenderText_Shaded
    sf = sdlttf.TTF_RenderText_Shaded(font, b"Hi there!", color, bgcolor)
    assert isinstance(sf.contents, surface.SDL_Surface)
    # Test TTF_RenderUTF8_Shaded
    teststr = u"Hï thère!".encode('utf-8')
    sf = sdlttf.TTF_RenderUTF8_Shaded(font, teststr, color, bgcolor)
    assert isinstance(sf.contents, surface.SDL_Surface)
    # Test TTF_RenderUNICODE_Shaded
    # NOTE: no unicode chars because number -> glyph lookup is os-dependent
    strarr = to_utf16(u"Hi there!")
    sf = sdlttf.TTF_RenderUNICODE_Shaded(font, strarr, color, bgcolor)
    assert isinstance(sf.contents, surface.SDL_Surface)
    # Test TTF_RenderGlyph_Shaded
    sf = sdlttf.TTF_RenderGlyph_Shaded(font, ord("A"), color, bgcolor)
    assert isinstance(sf.contents, surface.SDL_Surface)

@pytest.mark.skipif(sdlttf.dll.version < 2018, reason="not available")
def test_TTF_Render_Shaded_Wrapped(with_font):
    font = with_font
    color = SDL_Color(0, 0, 0, 255)
    bgcolor = SDL_Color(255, 255, 255)
    # Test TTF_RenderText_Shaded_Wrapped
    teststr = b"Hi there, this is a long line!"
    sf = sdlttf.TTF_RenderText_Shaded_Wrapped(font, teststr, color, bgcolor, 100)
    assert isinstance(sf.contents, surface.SDL_Surface)
    assert sf.contents.h > 30
    # Test TTF_RenderUTF8_Shaded_Wrapped
    teststr = u"Hï thère, this is a long line!".encode('utf-8')
    sf = sdlttf.TTF_RenderUTF8_Shaded_Wrapped(font, teststr, color, bgcolor, 100)
    assert isinstance(sf.contents, surface.SDL_Surface)
    assert sf.contents.h > 30
    # Test TTF_RenderUNICODE_Shaded_Wrapped
    # NOTE: no unicode chars because number -> glyph lookup is os-dependent
    strarr = to_utf16(u"Hi there, this is a long line!")
    sf = sdlttf.TTF_RenderUNICODE_Shaded_Wrapped(font, strarr, color, bgcolor, 100)
    assert isinstance(sf.contents, surface.SDL_Surface)
    assert sf.contents.h > 30

def test_TTF_Render_Blended(with_font):
    font = with_font
    color = SDL_Color(0, 0, 0, 255)
    # Test TTF_RenderText_Blended
    sf = sdlttf.TTF_RenderText_Blended(font, b"Hi there!", color)
    assert isinstance(sf.contents, surface.SDL_Surface)
    # Test TTF_RenderUTF8_Blended
    teststr = u"Hï thère!".encode('utf-8')
    sf = sdlttf.TTF_RenderUTF8_Blended(font, teststr, color)
    assert isinstance(sf.contents, surface.SDL_Surface)
    # Test TTF_RenderUNICODE_Blended
    # NOTE: no unicode chars because number -> glyph lookup is os-dependent
    strarr = to_utf16(u"Hi there!")
    sf = sdlttf.TTF_RenderUNICODE_Blended(font, strarr, color)
    assert isinstance(sf.contents, surface.SDL_Surface)
    # Test TTF_RenderGlyph_Blended
    sf = sdlttf.TTF_RenderGlyph_Blended(font, ord("A"), color)
    assert isinstance(sf.contents, surface.SDL_Surface)

def test_TTF_Render_Blended_Wrapped(with_font):
    font = with_font
    color = SDL_Color(0, 0, 0, 255)
    # Test TTF_RenderText_Blended_Wrapped
    teststr = b"Hi there, this is a long line!"
    sf = sdlttf.TTF_RenderText_Blended_Wrapped(font, teststr, color, 100)
    assert isinstance(sf.contents, surface.SDL_Surface)
    assert sf.contents.h > 30
    # Test TTF_RenderUTF8_Blended_Wrapped
    teststr = u"Hï thère, this is a long line!".encode('utf-8')
    sf = sdlttf.TTF_RenderUTF8_Blended_Wrapped(font, teststr, color, 100)
    assert isinstance(sf.contents, surface.SDL_Surface)
    assert sf.contents.h > 30
    # Test TTF_RenderUNICODE_Blended_Wrapped
    # NOTE: no unicode chars because number -> glyph lookup is os-dependent
    strarr = to_utf16(u"Hi there, this is a long line!")
    sf = sdlttf.TTF_RenderUNICODE_Blended_Wrapped(font, strarr, color, 100)
    assert isinstance(sf.contents, surface.SDL_Surface)
    assert sf.contents.h > 30

@pytest.mark.skipif(sdlttf.dll.version < 2018, reason="not available")
def test_TTF_RenderGlyph32(with_font):
    font = with_font
    color = SDL_Color(0, 0, 0)
    bgcolor = SDL_Color(255, 255, 255)
    # Test TTF_RenderGlyph32_Solid
    sf = sdlttf.TTF_RenderGlyph32_Solid(font, ord("A"), color)
    assert isinstance(sf.contents, surface.SDL_Surface)
    # Test TTF_RenderGlyph32_Shaded
    sf = sdlttf.TTF_RenderGlyph32_Shaded(font, ord("A"), color, bgcolor)
    assert isinstance(sf.contents, surface.SDL_Surface)
    # Test TTF_RenderGlyph32_Blended
    sf = sdlttf.TTF_RenderGlyph32_Blended(font, ord("A"), color)
    assert isinstance(sf.contents, surface.SDL_Surface)

@pytest.mark.skipif(sdlttf.dll.version < 2018, reason="not available")
def test_TTF_SetDirection(with_font):
    font = with_font
    # Define HarfBuzz direction constants
    HB_DIRECTION_LTR = 4
    HB_DIRECTION_TTB = 6
    # Only available if compiled with HarfBuzz
    major, minor, patch = c_int(0), c_int(0), c_int(0)
    sdlttf.TTF_GetHarfBuzzVersion(byref(major), byref(minor), byref(patch))
    if major.value == 0:
        pytest.skip("No HarfBuzz")
    # Try setting the script direction
    ret = sdlttf.TTF_SetDirection(HB_DIRECTION_LTR)
    assert ret == 0
    # Try changing the script direction to see if it has any effect
    w1, h1, w2, h2 = c_int(0), c_int(0), c_int(0), c_int(0)
    sdlttf.TTF_SizeText(font, b"Hi there!", byref(w1), byref(h1))
    ret = sdlttf.TTF_SetDirection(HB_DIRECTION_TTB)
    assert ret == 0
    sdlttf.TTF_SizeText(font, b"Hi there!", byref(w2), byref(h2))
    sdlttf.TTF_SetDirection(HB_DIRECTION_LTR) # Reset direction
    assert w1.value > w2.value
    assert h1.value < h2.value

@pytest.mark.skipif(sdlttf.dll.version < 2018, reason="not available")
def test_TTF_SetScript(with_font):
    # Only available if compiled with HarfBuzz
    major, minor, patch = c_int(0), c_int(0), c_int(0)
    sdlttf.TTF_GetHarfBuzzVersion(byref(major), byref(minor), byref(patch))
    if major.value == 0:
        pytest.skip("No HarfBuzz")
    # Try setting the script language to Arabic and back
    ret = sdlttf.TTF_SetScript(sdlttf.HB_TAG("A", "r", "a", "b"))
    assert ret == 0
    # NOTE: I have no clue how to write a proper test to see if this worked
    ret = sdlttf.TTF_SetScript(sdlttf.HB_TAG("Z", "y", "y", "y"))
    assert ret == 0

@pytest.mark.skipif(sdlttf.dll.version < 2014, reason="not available")
def test_TTF_GetFontKerningSizeGlyphs(with_font):
    font = with_font
    # NOTE: Test font (tuffy) has no kerning info, so retval is always 0
    sz = sdlttf.TTF_GetFontKerningSizeGlyphs(font, ord("A"), ord("B"))
    assert sz == 0

@pytest.mark.skipif(sdlttf.dll.version < 2018, reason="not available")
def test_TTF_GetFontKerningSizeGlyphs32(with_font):
    font = with_font
    # NOTE: Test font (tuffy) has no kerning info, so retval is always 0
    sz = sdlttf.TTF_GetFontKerningSizeGlyphs32(font, ord("A"), ord("B"))
    assert sz == 0

@pytest.mark.skipif(sdlttf.dll.version < 2018, reason="not available")
def test_TTF_GetSetFontSDF(with_font):
    font = with_font
    # Only available with FreeType >= 2.11.0
    major, minor, patch = c_int(0), c_int(0), c_int(0)
    sdlttf.TTF_GetFreeTypeVersion(byref(major), byref(minor), byref(patch))
    if minor.value < 11:
        pytest.skip("SDF not available")
    # Try enabling SDF rendering on a given font
    ret = sdlttf.TTF_SetFontSDF(font, SDL_TRUE)
    assert ret == 0
    assert sdlttf.TTF_GetFontSDF(font) == SDL_TRUE
    # Try disabling SDF rendering on a given font
    ret = sdlttf.TTF_SetFontSDF(font, SDL_FALSE)
    assert ret == 0
    assert sdlttf.TTF_GetFontSDF(font) == SDL_FALSE


def test_HB_TAG():
    test_scripts = {
        "Arab": 1098015074,
        "Mong": 1299148391,
        "Zyyy": 1517910393, # HB_SCRIPT_COMMON
        "Zzzz": 1517976186, # HB_SCRIPT_UNKNOWN
    }
    for script, expected in test_scripts.items():
        c1, c2, c3, c4 = script
        assert sdlttf.HB_TAG(c1, c2, c3, c4) == expected



from ctypes import POINTER, c_int
from .dll import _bind

__all__ = [
    # Enums
    "SDL_PowerState",
    "SDL_POWERSTATE_UNKNOWN", "SDL_POWERSTATE_ON_BATTERY",
    "SDL_POWERSTATE_NO_BATTERY", "SDL_POWERSTATE_CHARGING",
    "SDL_POWERSTATE_CHARGED",

    # Functions
    "SDL_GetPowerInfo"
]


SDL_PowerState = c_int

SDL_POWERSTATE_UNKNOWN = 0
SDL_POWERSTATE_ON_BATTERY = 1
SDL_POWERSTATE_NO_BATTERY = 2
SDL_POWERSTATE_CHARGING = 3
SDL_POWERSTATE_CHARGED = 4


SDL_GetPowerInfo = _bind("SDL_GetPowerInfo", [POINTER(c_int), POINTER(c_int)], SDL_PowerState)



import sys
import gc
import pytest
from ctypes import POINTER, byref, addressof

from sdl2 import ext as sdl2ext
from sdl2.render import (
    SDL_Renderer, SDL_CreateWindowAndRenderer, SDL_DestroyRenderer,
    SDL_Texture, SDL_CreateTexture
)
from sdl2.surface import SDL_CreateRGBSurface
from sdl2.video import SDL_Window, SDL_WINDOW_HIDDEN, SDL_DestroyWindow


class MSprite(sdl2ext.Sprite):

    def __init__(self, w=0, h=0):
        super(MSprite, self).__init__()
        self._size = w, h

    @property
    def size(self):
        return self._size


@pytest.fixture
def positions():
    test_positions = []
    for x in range(-50, 50, 5):
        for y in range(-50, 50, 5):
            test_positions.append((x, y))
    return test_positions

@pytest.fixture
def sizes():
    test_sizes = [(4, 4)]
    for w in range(16, 256, 16):
        for h in range(16, 256, 16):
            test_sizes.append((w, h))
    return test_sizes


class TestExtSprite(object):
    __tags__ = ["sdl", "sdl2ext"]

    def test_init(self):
        sprite = MSprite()
        assert isinstance(sprite, MSprite)
        assert isinstance(sprite, sdl2ext.Sprite)

    def test_position_xy(self, positions):
        sprite = MSprite()
        for x, y in positions:
            sprite.position = x, y
            assert sprite.position == (x, y)
            sprite.x = x + 1
            sprite.y = y + 1
            assert sprite.position == (x + 1, y + 1)

    def test_area(self, sizes):
        for w, h in sizes:
            sprite = MSprite(w, h)
            assert sprite.size == (w, h)
            assert sprite.area == (0, 0, w, h)
            sprite.position = w, h
            assert sprite.area == (w, h, 2 * w, 2 * h)


class TestExtSoftwareSprite(object):
    __tags__ = ["sdl", "sdl2ext"]

    def test_init(self, with_sdl):
        sf = SDL_CreateRGBSurface(0, 10, 10, 32, 0, 0, 0, 0)

        # Test with SDL_Surface
        sprite = sdl2ext.SoftwareSprite(sf.contents, False)
        assert addressof(sprite.surface) == addressof(sf.contents)
        assert not sprite.free

        # Test with SDL_Surface pointer
        sprite = sdl2ext.SoftwareSprite(sf, False)
        assert addressof(sprite.surface) == addressof(sf.contents)
        assert not sprite.free

        with pytest.raises(TypeError):
            sdl2ext.SoftwareSprite(None, True)

    def test___repr__(self, with_sdl):
        sf = SDL_CreateRGBSurface(0, 10, 10, 32, 0, 0, 0, 0)
        sprite = sdl2ext.SoftwareSprite(sf.contents, True)
        assert repr(sprite) == "SoftwareSprite(size=(10, 10), bpp=32)"

    def test_position_xy(self, with_sdl, positions):
        sf = SDL_CreateRGBSurface(0, 10, 10, 32, 0, 0, 0, 0)
        sprite = sdl2ext.SoftwareSprite(sf.contents, True)
        assert isinstance(sprite, sdl2ext.SoftwareSprite)
        assert sprite.position == (0, 0)
        for x, y in positions:
            sprite.position = x, y
            assert sprite.position == (x, y)
            sprite.x = x + 1
            sprite.y = y + 1
            assert sprite.position == (x + 1, y + 1)

    def test_size(self, with_sdl, sizes):
        for w, h in sizes:
            sf = SDL_CreateRGBSurface(0, w, h, 32, 0, 0, 0, 0)
            sprite = sdl2ext.SoftwareSprite(sf.contents, True)
            assert isinstance(sprite, sdl2ext.SoftwareSprite)
            assert sprite.size == (w, h)

    def test_area(self, with_sdl):
        sf = SDL_CreateRGBSurface(0, 10, 10, 32, 0, 0, 0, 0)
        sprite = sdl2ext.SoftwareSprite(sf.contents, True)
        assert sprite.area == (0, 0, 10, 10)
        sprite.position = 7, 3
        assert sprite.area == (7, 3, 17, 13)
        sprite.position = -22, 99
        assert sprite.area == (-22, 99, -12, 109)
        def setarea(s, v):
            s.area = v
        with pytest.raises(AttributeError):
            setarea(sprite, (1, 2, 3, 4))


class TestExtTextureSprite(object):
    __tags__ = ["sdl", "sdl2ext"]

    def test_init(self, with_sdl):
        window = POINTER(SDL_Window)()
        renderer = POINTER(SDL_Renderer)()
        SDL_CreateWindowAndRenderer(
            10, 10, SDL_WINDOW_HIDDEN, byref(window), byref(renderer)
        )
        tex = SDL_CreateTexture(renderer, 0, 0, 10, 10)
        assert isinstance(tex.contents, SDL_Texture)
        sprite = sdl2ext.TextureSprite(tex.contents)
        assert isinstance(sprite, sdl2ext.TextureSprite)
        SDL_DestroyRenderer(renderer)
        SDL_DestroyWindow(window)

    def test_position_xy(self, with_sdl, positions):
        window = POINTER(SDL_Window)()
        renderer = POINTER(SDL_Renderer)()
        SDL_CreateWindowAndRenderer(
            10, 10, SDL_WINDOW_HIDDEN, byref(window), byref(renderer)
        )
        tex = SDL_CreateTexture(renderer, 0, 0, 10, 10)
        assert isinstance(tex.contents, SDL_Texture)
        sprite = sdl2ext.TextureSprite(tex.contents)
        assert isinstance(sprite, sdl2ext.TextureSprite)
        assert sprite.position == (0, 0)
        for x, y in positions:
            sprite.position = x, y
            assert sprite.position == (x, y)
            sprite.x = x + 1
            sprite.y = y + 1
            assert sprite.position == (x + 1, y + 1)
        SDL_DestroyRenderer(renderer)
        SDL_DestroyWindow(window)

    def test_size(self, with_sdl, sizes):
        window = POINTER(SDL_Window)()
        renderer = POINTER(SDL_Renderer)()
        SDL_CreateWindowAndRenderer(
            10, 10, SDL_WINDOW_HIDDEN, byref(window), byref(renderer)
        )
        for w, h in sizes:
            tex = SDL_CreateTexture(renderer, 0, 0, w, h)
            assert isinstance(tex.contents, SDL_Texture)
            sprite = sdl2ext.TextureSprite(tex.contents)
            assert isinstance(sprite, sdl2ext.TextureSprite)
            assert sprite.size == (w, h)
            del sprite
        SDL_DestroyRenderer(renderer)
        SDL_DestroyWindow(window)

    def test_area(self, with_sdl):
        window = POINTER(SDL_Window)()
        renderer = POINTER(SDL_Renderer)()
        SDL_CreateWindowAndRenderer(
            10, 10, SDL_WINDOW_HIDDEN, byref(window), byref(renderer)
        )
        tex = SDL_CreateTexture(renderer, 0, 0, 10, 20)
        assert isinstance(tex.contents, SDL_Texture)
        sprite = sdl2ext.TextureSprite(tex.contents)
        assert isinstance(sprite, sdl2ext.TextureSprite)
        assert sprite.area == (0, 0, 10, 20)
        sprite.position = 7, 3
        assert sprite.area == (7, 3, 17, 23)
        sprite.position = -22, 99
        assert sprite.area == (-22, 99, -12, 119)
        def setarea(s, v):
            s.area = v
        with pytest.raises(AttributeError):
            setarea(sprite, (1, 2, 3, 4))
        SDL_DestroyRenderer(renderer)
        SDL_DestroyWindow(window)



from ctypes import Structure, POINTER, c_char_p, c_void_p, cast, sizeof
from .dll import _bind
from .dll import version as sdl_version

__all__ = [
    # Structs
    "SDL_Locale",

    # Functions
    "SDL_GetPreferredLocales"
]


class SDL_Locale(Structure):
    _fields_ = [("language", c_char_p),
                ("country", c_char_p)
                ]

    def __repr__(self):
        lang, country = (self.language, self.country)
        if type(lang) == bytes and type(lang) != str:
            lang = lang.decode('utf-8')
            country = country.decode('utf-8') if country else ''
        country_str = ", country={0}".format(country) if country else ''
        return "SDL_Locale(lang={0}{1})".format(lang, country_str)


_SDL_GetPreferredLocales = _bind("SDL_GetPreferredLocales", None, c_void_p, added='2.0.14')

def SDL_GetPreferredLocales():
    """The base SDL2 function returns a pointer to an array of locales with an
    arbitrary length, which is just plain awful to deal with in Python. This
    wrapper returns a Python list of locales instead.
    """
    locales = []
    loc_size = sizeof(SDL_Locale)
    p = _SDL_GetPreferredLocales()  # Get void pointer to first locale in array
    while True:
        loc = cast(p, POINTER(SDL_Locale))
        if not loc:
            # Break if loc is a null pointer
            break
        if loc.contents.language == None:
            # Array of locales is terminated by locale with null language
            break
        locales.append(loc.contents)
        p = p + loc_size  # Increment pointer to next locale in array
    return locales


import os
import sys
import pytest
import ctypes
from sdl2.stdinc import SDL_TRUE
from sdl2 import video, syswm, version, SDL_GetError

# Check if using dummy video driver
DRIVER_DUMMY = os.getenv("SDL_VIDEODRIVER", "") == "dummy"


@pytest.mark.skipif(DRIVER_DUMMY, reason="Doesn't work with dummy driver")
def test_SDL_GetWindowWMInfo(with_sdl):
    unix_wms = [
        syswm.SDL_SYSWM_X11, syswm.SDL_SYSWM_DIRECTFB,
        syswm.SDL_SYSWM_WAYLAND, syswm.SDL_SYSWM_KMSDRM, 
    ]
    platform_wms = {
        "windows": [syswm.SDL_SYSWM_WINDOWS],
        "darwin": [syswm.SDL_SYSWM_COCOA],
        "unixlike": unix_wms,
        "os2": [syswm.SDL_SYSWM_OS2],
    }
    # Create a window and retrieve the current window manager info
    window = video.SDL_CreateWindow(
        b"Test", 10, 10, 10, 10, video.SDL_WINDOW_HIDDEN
    )
    wminfo = syswm.SDL_SysWMinfo()
    version.SDL_VERSION(wminfo.version)
    ret = syswm.SDL_GetWindowWMInfo(window, ctypes.byref(wminfo))
    video.SDL_DestroyWindow(window)
    assert SDL_GetError() == b""
    assert ret == SDL_TRUE
    # Test window manager types for different platforms
    platform = sys.platform
    if platform in ("win32", "cygwin", "msys"):
        platform = "windows"
    elif "linux" in platform or "bsd" in platform:
        platform = "unixlike"
    if platform in platform_wms.keys():
        assert wminfo.subsystem in platform_wms[platform]



from ctypes import (
    c_char_p, c_wchar_p, c_int, c_ubyte, c_ushort, c_size_t, c_void_p,
    Structure, POINTER
)
from .dll import _bind
from .stdinc import Uint32, SDL_bool


__all__ = [
    # Structs
    "SDL_hid_device", "SDL_hid_device_info",

    # Functions
    "SDL_hid_init", "SDL_hid_exit", "SDL_hid_device_change_count",
    "SDL_hid_enumerate", "SDL_hid_free_enumeration",
    "SDL_hid_open", "SDL_hid_open_path",
    "SDL_hid_write", "SDL_hid_read_timeout", "SDL_hid_read",
    "SDL_hid_set_nonblocking",
    "SDL_hid_send_feature_report", "SDL_hid_get_feature_report",
    "SDL_hid_close",
    "SDL_hid_get_manufacturer_string", "SDL_hid_get_product_string",
    "SDL_hid_get_serial_number_string", "SDL_hid_get_indexed_string",
    "SDL_hid_ble_scan",
]


class SDL_hid_device(c_void_p):
    pass

class SDL_hid_device_info(Structure):
    pass

SDL_hid_device_info._fields_ = [
    ("path", c_char_p),
    ("vendor_id", c_ushort),
    ("product_id", c_ushort),
    ("serial_number", c_wchar_p),
    ("release_number", c_ushort),
    ("manufacturer_string", c_wchar_p),
    ("product_string", c_wchar_p),
    ("usage_page", c_ushort),
    ("usage", c_ushort),
    ("interface_number", c_int),
    ("interface_class", c_int),
    ("interface_subclass", c_int),
    ("interface_protocol", c_int),
    ("next", POINTER(SDL_hid_device_info))
]


SDL_hid_init = _bind("SDL_hid_init", None, c_int, added='2.0.18')
SDL_hid_exit = _bind("SDL_hid_exit", None, c_int, added='2.0.18')
SDL_hid_device_change_count = _bind("SDL_hid_device_change_count", None, Uint32, added='2.0.18')
SDL_hid_enumerate = _bind("SDL_hid_enumerate", [c_ushort, c_ushort], POINTER(SDL_hid_device_info), added='2.0.18')
SDL_hid_free_enumeration = _bind("SDL_hid_free_enumeration", [POINTER(SDL_hid_device_info)], None, added='2.0.18')
SDL_hid_open = _bind("SDL_hid_open", [c_ushort, c_ushort, c_wchar_p], POINTER(SDL_hid_device), added='2.0.18')
SDL_hid_open_path = _bind("SDL_hid_open_path", [c_char_p, c_int], POINTER(SDL_hid_device), added='2.0.18')

# NOTE: not sure if POINTER(c_ubyte) is right for this, need to test
SDL_hid_write = _bind("SDL_hid_write", [POINTER(SDL_hid_device), POINTER(c_ubyte), c_size_t], c_int, added='2.0.18')
SDL_hid_read_timeout = _bind("SDL_hid_read_timeout", [POINTER(SDL_hid_device), POINTER(c_ubyte), c_size_t, c_int], c_int, added='2.0.18')
SDL_hid_read = _bind("SDL_hid_read", [POINTER(SDL_hid_device), POINTER(c_ubyte), c_size_t], c_int, added='2.0.18')

SDL_hid_set_nonblocking = _bind("SDL_hid_set_nonblocking", [POINTER(SDL_hid_device), c_int], c_int, added='2.0.18')
SDL_hid_send_feature_report = _bind("SDL_hid_send_feature_report", [POINTER(SDL_hid_device), POINTER(c_ubyte), c_size_t], c_int, added='2.0.18')
SDL_hid_get_feature_report = _bind("SDL_hid_get_feature_report", [POINTER(SDL_hid_device), POINTER(c_ubyte), c_size_t], c_int, added='2.0.18')
SDL_hid_close = _bind("SDL_hid_close", [POINTER(SDL_hid_device)], None, added='2.0.18')
SDL_hid_get_manufacturer_string = _bind("SDL_hid_get_manufacturer_string", [POINTER(SDL_hid_device), c_wchar_p, c_size_t], c_int, added='2.0.18')
SDL_hid_get_product_string = _bind("SDL_hid_get_product_string", [POINTER(SDL_hid_device), c_wchar_p, c_size_t], c_int, added='2.0.18')
SDL_hid_get_serial_number_string = _bind("SDL_hid_get_serial_number_string", [POINTER(SDL_hid_device), c_wchar_p, c_size_t], c_int, added='2.0.18')
SDL_hid_get_indexed_string = _bind("SDL_hid_get_indexed_string", [POINTER(SDL_hid_device), c_int, c_wchar_p, c_size_t], c_int, added='2.0.18')
SDL_hid_ble_scan = _bind("SDL_hid_ble_scan", [SDL_bool], None, added='2.0.18')



import sys

__all__ = ["liangbarsky", "cohensutherland", "clipline", "point_on_line"]


def cohensutherland(left, top, right, bottom, x1, y1, x2, y2):
    """Clips a line to a rectangular area.

    This implements the Cohen-Sutherland line clipping algorithm. ``left``,
    ``top``, ``right`` and ``bottom`` define the bounds of the clipping area,
    by which the line from ``(x1, y1)`` to ``(x2, y2)`` will be clipped.

    Args:
        left (int): The left boundary of the clipping area.
        top (int): The top boundary of the clipping area.
        right (int): The right boundary of the clipping area.
        bottom (int): The bottom boundary of the clipping area.
        x1 (int): The x-coordinate of the starting point of the line.
        y1 (int): The y-coordinate of the starting point of the line.
        x2 (int): The x-coordinate of the end point of the line.
        y2 (int): The y-coordinate of the end point of the line.

    Returns:
        tuple: The start and end coordinates of the clipped line in the form
        ``(cx1, cy1, cx2, cy2)``. If the line does not intersect with the
        rectangular clipping area, all 4 values will be ``None``.

    """
    LEFT, RIGHT, LOWER, UPPER = 1, 2, 4, 8

    if bottom > top:
        bottom, top = (top, bottom)

    def _getclip(xa, ya):
        p = 0
        if xa < left:
            p = LEFT
        elif xa > right:
            p = RIGHT
        if ya < bottom:
            p |= LOWER
        elif ya > top:
            p |= UPPER
        return p

    k1 = _getclip(x1, y1)
    k2 = _getclip(x2, y2)
    while (k1 | k2) != 0:
        if (k1 & k2) != 0:
            return None, None, None, None
        opt = k2 if k2 > k1 else k1
        if opt & UPPER:
            x = x1 + (x2 - x1) * (1.0 * (top - y1)) / (y2 - y1)
            y = top
        elif opt & LOWER:
            x = x1 + (x2 - x1) * (1.0 * (bottom - y1)) / (y2 - y1)
            y = bottom
        elif opt & RIGHT:
            y = y1 + (y2 - y1) * (1.0 * (right - x1)) / (x2 - x1)
            x = right
        elif opt & LEFT:
            y = y1 + (y2 - y1) * (1.0 * (left - x1)) / (x2 - x1)
            x = left
        else:
            # this should not happen
            raise RuntimeError("invalid clipping state")

        if opt == k1:
            x1, y1 = x, y
            k1 = _getclip(x1, y1)
        else:
            x2, y2 = x, y
            k2 = _getclip(x2, y2)

    return x1, y1, x2, y2


def liangbarsky(left, top, right, bottom, x1, y1, x2, y2):
    """Clips a line to a rectangular area.

    This implements the Liang-Barsky line clipping algorithm. ``left``,
    ``top``, ``right`` and ``bottom`` define the bounds of the clipping area,
    by which the line from ``(x1, y1)`` to ``(x2, y2)`` will be clipped.

    Args:
        left (int): The left boundary of the clipping area.
        top (int): The top boundary of the clipping area.
        right (int): The right boundary of the clipping area.
        bottom (int): The bottom boundary of the clipping area.
        x1 (int): The x-coordinate of the starting point of the line.
        y1 (int): The y-coordinate of the starting point of the line.
        x2 (int): The x-coordinate of the end point of the line.
        y2 (int): The y-coordinate of the end point of the line.

    Returns:
        tuple: The start and end coordinates of the clipped line in the form
        ``(cx1, cy1, cx2, cy2)``. If the line does not intersect with the
        rectangular clipping area, all 4 values will be ``None``.

    """
    dx = x2 - x1 * 1.0
    dy = y2 - y1 * 1.0
    dt0, dt1 = 0.0, 1.0
    xx1 = x1
    yy1 = y1

    if bottom > top:
        bottom, top = (top, bottom)

    checks = ((-dx, x1 - left),
              (dx, right - x1),
              (-dy, y1 - bottom),
              (dy, top - y1))

    for p, q in checks:
        if p == 0 and q < 0:
            return None, None, None, None
        if p != 0:
            dt = q / (p * 1.0)
            if p < 0:
                if dt > dt1:
                    return None, None, None, None
                dt0 = max(dt0, dt)
            else:
                if dt < dt0:
                    return None, None, None, None
                dt1 = min(dt1, dt)

    if dt0 > 0:
        x1 += dt0 * dx
        y1 += dt0 * dy
    if dt1 < 1:
        x2 = xx1 + dt1 * dx
        y2 = yy1 + dt1 * dy

    return x1, y1, x2, y2


def clipline(l, t, r, b, x1, y1, x2, y2, method='liangbarsky'):
    """Clips a line to a rectangular area using a given method.

    Args:
        l (int): The left boundary of the clipping area.
        t (int): The top boundary of the clipping area.
        r (int): The right boundary of the clipping area.
        b (int): The bottom boundary of the clipping area.
        x1 (int): The x-coordinate of the starting point of the line.
        y1 (int): The y-coordinate of the starting point of the line.
        x2 (int): The x-coordinate of the end point of the line.
        y2 (int): The y-coordinate of the end point of the line.
        method (str, optional): The method to use for clipping lines, can be
            either 'cohensutherland' or 'liangbarsky'. Defaults to liangbarsky.

    Returns:
        tuple: The start and end coordinates of the clipped line in the form
        ``(cx1, cy1, cx2, cy2)``. If the line does not intersect with the
        rectangular clipping area, all 4 values will be ``None``.

    """
    if method == 'cohensutherland':
        return cohensutherland(l, t, r, b, x1, y1, x2, y2)
    elif method == 'liangbarsky':
        return liangbarsky(l, t, r, b, x1, y1, x2, y2)
    else:
        raise ValueError("Unknown clipping method '{0}'".format(method))


def point_on_line(p1, p2, point):
    """Checks if a point falls along a given line segment.

    Args:
        p1 (tuple): The (x, y) coordinates of the starting point of the line.
        p2 (tuple): The (x, y) coordinates of the end point of the line.
        point (tuple): The (x, y) coordinates to test against the line.
    
    Returns:
        bool: ``True`` if the point falls along the line segment, otherwise
        ``False``.

    """
    x1, y1 = p1
    x2, y2 = p2
    px, py = point
    det = (py - y1) * (x2 - x1) - (px - x1) * (y2 - y1)
    if abs(det) > sys.float_info.epsilon:
        return False
    return (min(x1, x2) <= px <= max(x1, x2) and
            min(y1, y2) <= py <= max(y1, y2))



# -*- coding: utf-8 -*-
import gc
import pytest
from sdl2 import ext as sdl2ext
from sdl2.ext.compat import byteify
from sdl2.ext.pixelaccess import pixels2d
from sdl2.ext.surface import _create_surface
from sdl2 import surface, pixels, rwops, SDL_ClearError, SDL_GetError

_HASSDLTTF = True
try:
    from sdl2 import sdlttf
except ImportError:
    _HASSDLTTF = False


RESOURCES = sdl2ext.Resources(__file__, "resources")
RED_RGBA = (255, 0, 0, 255)
FONTMAP = ["0123456789",
           "ABCDEFGHIJ",
           "KLMNOPQRST",
           "UVWXYZ    ",
           "abcdefghij",
           "klmnopqrst",
           "uvwxyz    ",
           ",;.:!?-+()"
           ]


@pytest.fixture(scope="module")
def with_sdl():
    sdl2ext.init()
    yield
    sdl2ext.quit()

@pytest.fixture()
def with_font_ttf(with_sdl):
    if _HASSDLTTF:
        SDL_ClearError()
        fontpath = RESOURCES.get_path("tuffy.ttf")
        font = sdl2ext.FontTTF(fontpath, 20, RED_RGBA)
        assert SDL_GetError() == b""
        assert font
        yield font
        font.close()
        gc.collect()


class TestBitmapFont(object):
    __tags__ = ["sdl", "sdl2ext"]

    @classmethod
    def setup_class(cls):
        try:
            sdl2ext.init()
        except sdl2ext.SDLError:
            raise pytest.skip('Video subsystem not supported')

    @classmethod
    def teardown_class(cls):
        sdl2ext.quit()

    def setup_method(self):
        SDL_ClearError()

    def teardown_method(self):
        gc.collect()

    def test_init(self):
        # Initialize surface and sprite for tests
        fontpath = byteify(RESOURCES.get_path("font.bmp"), "utf-8")
        sf = surface.SDL_LoadBMP(fontpath)
        sprite = sdl2ext.SoftwareSprite(sf.contents, True)

        # Try SoftwareSprite surface
        font = sdl2ext.BitmapFont(sprite, (32, 32), FONTMAP)
        assert font.size == (32, 32)

        # Try SDL_Surface surface
        font = sdl2ext.BitmapFont(sf.contents, (32, 32), FONTMAP)
        assert font.size == (32, 32)

        # Try SDL_Surface pointer surface
        font = sdl2ext.BitmapFont(sf, (32, 32), FONTMAP)
        assert font.size == (32, 32)

        # Try loading a font directly from a .bmp
        font = sdl2ext.BitmapFont(fontpath, (32, 32), FONTMAP)
        assert font.size == (32, 32)

        # Test use of default fontmap and inferred character size
        font = sdl2ext.BitmapFont(sf)
        assert font.size == (32, 32)

        # Try invalid surface type
        with pytest.raises(TypeError):
            sdl2ext.BitmapFont([], (32, 32), FONTMAP)

    def test_render(self):
        # Initialize font and BitmapFont for tests
        fontpath = byteify(RESOURCES.get_path("font.bmp"), "utf-8")
        sf = surface.SDL_LoadBMP(fontpath)
        font = sdl2ext.BitmapFont(sf.contents, (32, 32), FONTMAP)

        # Try rendering some text
        msg = "hello there!"
        text = font.render(msg)
        assert isinstance(text, sdl2ext.SoftwareSprite)
        assert text.size[0] == 32 * len(msg)

        # Test exception for missing glyph
        with pytest.raises(ValueError):
            font.render("this_should_fail")

    def test_render_text(self):
        # Initialize BitmapFont and dummy RGB888 surface for tests
        fontpath = byteify(RESOURCES.get_path("font.bmp"), "utf-8")
        font = sdl2ext.BitmapFont(fontpath)
        rgb_surf = _create_surface(size=(320, 256), fmt="RGB888")
        font_rgb = sdl2ext.BitmapFont(rgb_surf)

        # Try rendering some text
        msg = "hello there!"
        text = font.render_text(msg)
        assert isinstance(text, surface.SDL_Surface)
        assert text.w == 32 * len(msg)
        assert text.h == 32
        surface.SDL_FreeSurface(text)

        # Try rendering with a different line height
        text = font.render_text(msg, line_h=40)
        assert isinstance(text, surface.SDL_Surface)
        assert text.w == 32 * len(msg)
        assert text.h == 40
        surface.SDL_FreeSurface(text)

        # Make sure ARGB converion works
        text = font_rgb.render_text(msg)
        assert isinstance(text, surface.SDL_Surface)
        assert text.w == 32 * len(msg)
        assert text.h == 32
        assert text.format.contents.format == pixels.SDL_PIXELFORMAT_ARGB8888
        surface.SDL_FreeSurface(text)

        # Try rendering without ARGB conversion
        text = font_rgb.render_text(msg, as_argb=False)
        assert isinstance(text, surface.SDL_Surface)
        assert text.w == 32 * len(msg)
        assert text.h == 32
        assert text.format.contents.format == pixels.SDL_PIXELFORMAT_RGB888
        surface.SDL_FreeSurface(text)

        # Test multiline rendering
        msg = "hello\nthere!"
        text = font.render_text(msg)
        assert isinstance(text, surface.SDL_Surface)
        assert text.w == 32 * 6
        assert text.h == 64
        surface.SDL_FreeSurface(text)

        # Test exception for missing glyph
        with pytest.raises(ValueError):
            font.render_text("this_should_fail")

    def test_render_on(self):
        np = pytest.importorskip("numpy", reason="numpy module is not available")
        # Initialize BitmapFont for tests
        fontpath = byteify(RESOURCES.get_path("font.bmp"), "utf-8")
        font = sdl2ext.BitmapFont(fontpath)

        # Try rendering some text
        target = _create_surface(size=(32*5, 32))
        view = pixels2d(target, transpose=False)
        mid_row = view[16, :].copy()
        outrect = font.render_on(target, "TEST!")
        assert not np.all(mid_row == view[16, :]) # ensure surface changed
        assert outrect == (0, 0, 32*5, 32)

        # Try rendering some text with an offset
        target2 = _create_surface(size=(32*5, 32))
        view2 = pixels2d(target2, transpose=False)
        outrect2 = font.render_on(target2, "TEST!", offset=(5, 0))
        assert not np.all(mid_row == view2[16, :]) # ensure surface changed
        assert not np.all(view[16, :] == view2[16, :]) # ensure offset worked
        assert outrect2 == (5, 0, 32*5 + 5, 32)

        surface.SDL_FreeSurface(target)
        surface.SDL_FreeSurface(target2)

        # Test exception for missing glyph
        with pytest.raises(ValueError):
            font.render_on(target, "%nope")

    def test_remap(self):
        # Initialize BitmapFont for tests
        fontpath = byteify(RESOURCES.get_path("font.bmp"), "utf-8")
        font = sdl2ext.BitmapFont(fontpath)
        msg = "hello there!"

        # Remap the l to be narrower and try rendering
        font.remap("l", 32, 160, 12, 32)
        text = font.render_text(msg)
        assert text.w == (32 * len(msg) - 40)
        assert text.h == 32

        surface.SDL_FreeSurface(text)

        # Test exceptions on bad input
        with pytest.raises(ValueError):
            font.remap("hi", 4, 4, 4, 4)
        with pytest.raises(ValueError):
            font.remap("h", 10, 32, 32, 0)
        with pytest.raises(ValueError):
            font.remap("h", 32, 1000, 32, 32)

    def test_contains(self):
        # Initialize BitmapFont for tests
        fontpath = byteify(RESOURCES.get_path("font.bmp"), "utf-8")
        font = sdl2ext.BitmapFont(fontpath)

        for ch in "abcde12345.-,+":
            assert font.contains(ch)
        del font.offsets[" "]
        for ch in u" äöüß":
            assert not font.contains(ch)

    def test_can_render(self):
        sf = surface.SDL_LoadBMP(byteify(RESOURCES.get_path("font.bmp"),
                                         "utf-8"))
        assert isinstance(sf.contents, surface.SDL_Surface)
        font = sdl2ext.BitmapFont(sf, (32, 32), FONTMAP)
        assert isinstance(font, sdl2ext.BitmapFont)

        assert font.can_render("text")
        assert font.can_render("473285435hfsjadfhriuewtrhefd")
        assert not font.can_render("testä")



@pytest.mark.skipif(not _HASSDLTTF, reason="SDL_TTF library not available")
class TestFontTTF(object):
    __tags__ = ["sdl", "sdl2ext"]

    def test_init(self, with_sdl):
        # Try opening and closing a font
        fontpath = RESOURCES.get_path("tuffy.ttf")
        font = sdl2ext.FontTTF(fontpath, 20, RED_RGBA)
        assert SDL_GetError() == b""
        assert font
        font.close()

        # Try opening a font with size specified in pt
        font = sdl2ext.FontTTF(fontpath, "20pt", RED_RGBA)
        assert SDL_GetError() == b""
        assert font
        font.close()

        # Try opening a font with size specified in pixels
        font = sdl2ext.FontTTF(fontpath, "20px", RED_RGBA)
        assert SDL_GetError() == b""
        assert font
        
        # Try opening a font with a custom set of height chars
        chars = "aeiou"
        font2 = sdl2ext.FontTTF(fontpath, "20px", RED_RGBA, height_chars=chars)
        assert SDL_GetError() == b""
        assert font
        assert font._parse_size("20px") != font2._parse_size("20px")
        font.close()
        font2.close()

        # Test exception on invalid font path
        with pytest.raises(IOError):
            sdl2ext.FontTTF("missing.ttf", 16, RED_RGBA)
        
        # Text exception on invalid size
        with pytest.raises(ValueError):
            sdl2ext.FontTTF(fontpath, 0, RED_RGBA)

        # Test exception on missing height chars
        with pytest.raises(RuntimeError):
            sdl2ext.FontTTF(fontpath, 20, RED_RGBA, height_chars=u"的是不")

    def test_init_rw(self, with_sdl):
        # Try opening and closing a font
        fontpath = RESOURCES.get_path("tuffy.ttf")
        font_file = open(fontpath, "rb")
        font_rw = rwops.rw_from_object(font_file)
        font = sdl2ext.FontTTF(font_rw, 20, RED_RGBA)
        assert SDL_GetError() == b""
        assert font
        font.close()

    def test_get_ttf_font(self, with_font_ttf):
        font = with_font_ttf
        ttf_font = font.get_ttf_font()
        assert isinstance(ttf_font, sdlttf.TTF_Font)

        # Test exception for missing style
        with pytest.raises(ValueError):
            font.get_ttf_font("small")

    def test_add_style(self, with_font_ttf):
        font = with_font_ttf

        # Add some new styles
        font.add_style('large', '40px', RED_RGBA)
        font.add_style('red', 20, pixels.SDL_Color(255, 0, 0))
        font.add_style('red_on_white', 20, (255, 0, 0), (255, 255, 255))

        # Test exception for existing style name
        with pytest.raises(ValueError):
            font.add_style('large', '50px', RED_RGBA)

        # Test exception for invalid size unit
        with pytest.raises(ValueError):
            font.add_style('unit_err', '10cm', RED_RGBA)

        # Test exception for non-integer size
        with pytest.raises(ValueError):
            font.add_style('float_err', '10.5pt', RED_RGBA)

    def test_render_text(self, with_font_ttf):
        font = with_font_ttf

        # Try rendering some text
        msg = "hello there!"
        text = font.render_text(msg)
        assert SDL_GetError() == b""
        assert isinstance(text, surface.SDL_Surface)
        assert text.format.contents.format == pixels.SDL_PIXELFORMAT_ARGB8888

        # Test multiline rendering
        msg = "hello\nthere!"
        text2 = font.render_text(msg)
        assert SDL_GetError() == b""
        assert isinstance(text2, surface.SDL_Surface)
        assert text2.h > text.h

        # Test custom line height
        msg = "hello\nthere!"
        text3 = font.render_text(msg, line_h=100)
        assert SDL_GetError() == b""
        assert isinstance(text3, surface.SDL_Surface)
        assert text3.h > text2.h
        surface.SDL_FreeSurface(text3)

        # Test custom line height as a string (in px)
        msg = "hello\nthere!"
        text3 = font.render_text(msg, line_h='100px')
        assert SDL_GetError() == b""
        assert isinstance(text3, surface.SDL_Surface)
        assert text3.h > text2.h
        surface.SDL_FreeSurface(text3)

        # Test custom line height as a percentage
        msg = "hello\nthere!"
        text3 = font.render_text(msg, line_h='110%')
        assert SDL_GetError() == b""
        assert isinstance(text3, surface.SDL_Surface)
        assert text3.h > text2.h
        surface.SDL_FreeSurface(text3)

        # Test wrap width
        msg = "hello there! This is a very long line of text."
        text3 = font.render_text(msg, width=200)
        assert SDL_GetError() == b""
        assert isinstance(text3, surface.SDL_Surface)
        assert text3.h > text.h
        surface.SDL_FreeSurface(text3)

        # Test custom font style (no bg color)
        msg = "hello there!"
        font.add_style("blue", 30, (0, 0, 255, 255))
        text3 = font.render_text(msg, "blue")
        assert SDL_GetError() == b""
        assert isinstance(text3, surface.SDL_Surface)
        surface.SDL_FreeSurface(text3)

        # Test custom font style (w/ bg color)
        msg = "hello there!"
        font.add_style("red_on_white", 30, RED_RGBA, (255, 255, 255))
        text3 = font.render_text(msg, "red_on_white")
        assert SDL_GetError() == b""
        assert isinstance(text3, surface.SDL_Surface)
        surface.SDL_FreeSurface(text3)

        # Test font alignment (not extensive)
        msg = "hello there!\nThis is a very long line of\ntext."
        for align in ["left", "center", "right"]:
            tmp = font.render_text(msg, align=align)
            assert SDL_GetError() == b""
            assert isinstance(text3, surface.SDL_Surface)
            surface.SDL_FreeSurface(tmp)

        surface.SDL_FreeSurface(text)
        surface.SDL_FreeSurface(text2)

        # Test exception for empty string
        with pytest.raises(ValueError):
            font.render_text("")

        # Test exception for missing style
        with pytest.raises(ValueError):
            font.render_text(msg, "small")

        # Test exceptions for bad line height
        with pytest.raises(ValueError):
            font.render_text(msg, line_h=12.4)
        with pytest.raises(ValueError):
            font.render_text(msg, line_h=0)
        with pytest.raises(ValueError):
            font.render_text(msg, line_h='-50%')
        with pytest.raises(ValueError):
            font.render_text(msg, line_h='100pt')

        # Test exception for bad alignment
        with pytest.raises(ValueError):
            font.render_text(msg, align="flush")

    def test_close(self, with_font_ttf):
        font = with_font_ttf
        font.close()

        # Test that you can close the font multiple times
        font.close()

        # Make sure you can't use font after closing it
        with pytest.raises(RuntimeError):
            font.get_ttf_font()
        with pytest.raises(RuntimeError):
            font.add_style('large', 40, RED_RGBA)
        with pytest.raises(RuntimeError):
            font.render_text("hello!")
        with pytest.raises(RuntimeError):
            font.contains("A")
        with pytest.raises(RuntimeError):
            font.family_name
        with pytest.raises(RuntimeError):
            font.style_name
        with pytest.raises(RuntimeError):
            font.is_fixed_width

    def test_contains(self, with_font_ttf):
        font = with_font_ttf
        for ch in "abcde12345":
            assert font.contains(ch)
        for ch in u"的是不":
            assert not font.contains(ch)

    def test_family_name(self, with_font_ttf):
        font = with_font_ttf
        name = font.family_name
        assert name == None or isinstance(name, str)

    def test_style_name(self, with_font_ttf):
        font = with_font_ttf
        name = font.style_name
        assert name == None or isinstance(name, str)

    def test_is_fixed_width(self, with_font_ttf):
        font = with_font_ttf
        assert font.is_fixed_width == False


@pytest.mark.skipif(not _HASSDLTTF, reason="SDL_TTF library not available")
class TestFontManager(object):
    __tags__ = ["sdl", "sdl2ext"]

    @classmethod
    def setup_class(cls):
        try:
            sdl2ext.init()
        except sdl2ext.SDLError:
            raise pytest.skip('Video subsystem not supported')

    @classmethod
    def teardown_class(cls):
        sdl2ext.quit()

    def setup_method(self):
        SDL_ClearError()

    def teardown_method(self):
        gc.collect()

    def test_init(self):
        fm = sdl2ext.FontManager(RESOURCES.get_path("tuffy.ttf"),
                                 bg_color=(100, 0, 0))
        assert isinstance(fm, sdl2ext.FontManager)
        assert fm.default_font == "tuffy"
        assert fm.size == 16
        assert fm.bg_color == sdl2ext.Color(100, 0, 0, 0)

    def test_default_font(self):
        fontpath = RESOURCES.get_path("tuffy.ttf")
        fm = sdl2ext.FontManager(fontpath)
        assert fm.default_font == "tuffy"
        assert fm.size == 16
        with pytest.raises(ValueError):
            fm.default_font = "Inexistent Alias"
        fm.add(fontpath, alias = "tuffy.copy", size = 10)
        fm.default_font = "tuffy.copy"
        fm.size = 10
        assert fm.default_font == "tuffy.copy"
        assert fm.size == 10
        fm.default_font = "tuffy.copy"
        fm.size = 16
        assert fm.default_font == "tuffy.copy"
        assert fm.size == 16

    def test_add(self):
        fm = sdl2ext.FontManager(RESOURCES.get_path("tuffy.ttf"))
        assert "tuffy" in fm.aliases
        assert "tuffy" in fm.fonts
        assert 16 in fm.fonts["tuffy"]
        assert isinstance(fm.fonts["tuffy"][16].contents, sdlttf.TTF_Font)

        # Do some metrics tests
        # NOTE: Ascent & other font metrics changed in FreeType 2.10, so we 
        # test against both < 2.10 and >= 2.10 values
        font = fm.fonts["tuffy"][16]
        assert sdlttf.TTF_FontAscent(font) in [13, 16]
        fm.add(RESOURCES.get_path("tuffy.ttf"), size=12)
        font = fm.fonts["tuffy"][12]
        assert sdlttf.TTF_FontAscent(font) in [10, 12]

        with pytest.raises(IOError):
            fm.add("inexistent.ttf")
        # I don't find a scenario raising a TTF_Error.
        # self.assertRaises(sdl2ext.SDLError, fm.add, "resources/tuffy.ttf",
        #                   size=-1)

        # Close the font manager and add a new font
        fm.close()
        fm.add(RESOURCES.get_path("tuffy.ttf"), size=12)
        assert isinstance(fm.fonts["tuffy"][12].contents, sdlttf.TTF_Font)

    def test_close(self):
        fm = sdl2ext.FontManager(RESOURCES.get_path("tuffy.ttf"))
        fm.add(RESOURCES.get_path("tuffy.ttf"), size=20)
        fm.add(RESOURCES.get_path("tuffy.ttf"), alias="Foo", size=10)
        fm.close()
        assert fm.fonts == {}
        # How to make sure TTF_CloseFont was called on each loaded font?

    def test_render(self):
        fm = sdl2ext.FontManager(RESOURCES.get_path("tuffy.ttf"))
        text_surf = fm.render("text")
        assert isinstance(text_surf, surface.SDL_Surface)
        assert text_surf.w > 1

        text_surf = fm.render("text", size=10)
        assert isinstance(text_surf, surface.SDL_Surface)

        text_surf = fm.render("""
text long enough to have it wrapped at 100 px width.""", size=20, width=100)
        assert isinstance(text_surf, surface.SDL_Surface)
        assert text_surf.w > 1
        assert text_surf.w == 100
        with pytest.raises(KeyError):
            fm.render("text", alias="inexistent")



from ctypes import POINTER, c_int, c_void_p
from .dll import _bind
from .video import SDL_Window

__all__ = [
    # Opaque Types
    "SDL_MetalView",
    
    # Functions
    "SDL_Metal_CreateView", "SDL_Metal_DestroyView", "SDL_Metal_GetLayer",
    "SDL_Metal_GetDrawableSize"
]

# NOTE: These functions are currently untested, but proper usage likely involves
# the use of pyobjc to create an NSView from the created SDL_MetalView.


#SDL_MetalView = c_void_p
class SDL_MetalView(c_void_p):
    pass

SDL_Metal_CreateView = _bind("SDL_Metal_CreateView", [POINTER(SDL_Window)], SDL_MetalView, added='2.0.12')
SDL_Metal_DestroyView = _bind("SDL_Metal_DestroyView", [SDL_MetalView], None, added='2.0.12')
SDL_Metal_GetLayer = _bind("SDL_Metal_GetLayer", [SDL_MetalView], c_void_p, added='2.0.14')
SDL_Metal_GetDrawableSize = _bind("SDL_Metal_GetDrawableSize", [POINTER(SDL_Window), POINTER(c_int), POINTER(c_int)], None, added='2.0.14')



# pytest configuration file
import os
import gc
import pytest
import sdl2
from sdl2 import ext as sdl2ext

# A flag to skip annoying audiovisual tests (e.g. window minimize).
# Defaults to True unless an environment variable is explicitly set.
SKIP_ANNOYING = os.getenv("PYSDL2_ALL_TESTS", "0") == "0"

@pytest.fixture(scope="module")
def with_sdl():
    sdl2.SDL_ClearError()
    ret = sdl2.SDL_Init(sdl2.SDL_INIT_VIDEO | sdl2.SDL_INIT_TIMER)
    assert sdl2.SDL_GetError() == b""
    assert ret == 0
    yield
    sdl2.SDL_Quit()

@pytest.fixture(autouse=True)
def sdl_cleanup():
    sdl2.SDL_ClearError()
    yield
    sdl2.SDL_ClearError()
    gc.collect()



import pytest
import sdl2
from sdl2 import locale


@pytest.mark.skipif(sdl2.dll.version < 2014, reason="not available")
def test_SDL_GetPreferredLocales():
    locales = locale.SDL_GetPreferredLocales()
    assert type(locales) == list
    if len(locales):
        assert all([type(i) == locale.SDL_Locale for i in locales])
        assert len(locales[0].language)



import sys
import pytest
from sdl2.surface import SDL_CreateRGBSurface, SDL_FreeSurface
from sdl2.rect import SDL_Rect
from sdl2.error import SDL_GetError
from sdl2.ext.color import Color, COLOR
from sdl2.ext.surface import _create_surface
from sdl2 import ext as sdl2ext

try:
    import numpy
    _HASNUMPY = True
except:
    _HASNUMPY = False


@pytest.fixture
def testsurf(with_sdl):
    sf = SDL_CreateRGBSurface(0, 10, 10, 32, 0, 0, 0, 0)
    assert SDL_GetError() == b""
    yield sf
    SDL_FreeSurface(sf)


@pytest.mark.skipif(not _HASNUMPY, reason="pixels3d requires numpy module")
def test_fill(testsurf):
    # Initialize colour and surface/view
    WHITE = (255, 255, 255)
    BLACK = (0, 0, 0)
    sf = testsurf
    view = sdl2ext.pixels3d(sf.contents, False)

    # Test with no provided fill area
    sdl2ext.fill(sf.contents, WHITE, None)
    assert all(x == 255 for x in view[0][0][:3])
    assert all(x == 255 for x in view[-1][-1][:3])

    # Test with SDL_Rect fill area
    sdl2ext.fill(sf.contents, BLACK, None)  # reset surface
    r = SDL_Rect(0, 0, 5, 5)
    sdl2ext.fill(sf.contents, WHITE, r)
    assert all(x == 255 for x in view[0][0][:3])
    assert all(x == 255 for x in view[4][4][:3])
    assert all(x == 0 for x in view[-1][-1][:3])

    # Test with tuple fill area
    sdl2ext.fill(sf.contents, BLACK, None)  # reset surface
    r = (5, 5, 5, 5)
    sdl2ext.fill(sf.contents, WHITE, r)
    assert all(x == 0 for x in view[4][4][:3])
    assert all(x == 255 for x in view[5][5][:3])
    assert all(x == 255 for x in view[-1][-1][:3])

    # Test with multiple fill areas
    sdl2ext.fill(sf.contents, BLACK, None)  # reset surface
    rects = [(0, 0, 10, 5), SDL_Rect(0, 0, 3, 10), (7, 7, 3, 10)]
    sdl2ext.fill(sf.contents, WHITE, rects)
    assert all(x == 255 for x in view[0][0][:3])
    assert all(x == 255 for x in view[0][-1][:3])
    assert all(x == 255 for x in view[-1][-1][:3])
    assert all(x == 0 for x in view[-1][4][:3])

    # Test exception on bad input
    with pytest.raises(ValueError):
        sdl2ext.fill(sf.contents, WHITE, (1, 2, 3))

@pytest.mark.skipif(not _HASNUMPY, reason="pixels3d requires numpy module")
def test_line(testsurf):
    # Initialize colour and surface/view
    WHITE = (255, 255, 255)
    BLACK = (0, 0, 0)
    sf = testsurf
    view = sdl2ext.pixels3d(sf.contents, False)

    # Test with a single straight line
    sdl2ext.line(sf.contents, WHITE, (0, 0, 11, 0))
    assert all(x == 255 for x in view[0][0][:3])
    assert all(x == 255 for x in view[0][-1][:3])
    assert all(x == 0 for x in view[1][0][:3])

    # Test specifying line width
    sdl2ext.line(sf.contents, WHITE, (0, 5, 11, 5), width=4)
    assert all(x == 255 for x in view[3][0][:3])
    assert all(x == 255 for x in view[6][-1][:3])
    assert all(x == 0 for x in view[2][0][:3])

    # Test with a single diagonal line
    sdl2ext.fill(sf.contents, BLACK, None)  # reset surface
    sdl2ext.line(sf.contents, WHITE, (1, 1, 9, 9))
    assert all(x == 255 for x in view[2][2][:3])
    assert all(x == 255 for x in view[3][3][:3])
    assert all(x == 0 for x in view[4][6][:3])

    # Test with multiple lines
    lines = [(0, 0, 0, 10), (0, 0, 10, 0), (0, 0, 10, 10)]
    sdl2ext.fill(sf.contents, BLACK, None)  # reset surface
    sdl2ext.line(sf.contents, WHITE, lines)
    assert all(x == 255 for x in view[0][-1][:3])
    assert all(x == 255 for x in view[-1][0][:3])
    assert all(x == 255 for x in view[-1][-1][:3])
    assert all(x == 0 for x in view[1][5][:3])

    # Test with multiple lines (old weird method)
    lines = (0, 0, 0, 10, 0, 0, 10, 0, 0, 0, 10, 10)
    sdl2ext.fill(sf.contents, BLACK, None)  # reset surface
    sdl2ext.line(sf.contents, WHITE, lines)
    assert all(x == 255 for x in view[0][-1][:3])
    assert all(x == 255 for x in view[-1][0][:3])
    assert all(x == 255 for x in view[-1][-1][:3])
    assert all(x == 0 for x in view[1][5][:3])

    # Test exception on bad input
    with pytest.raises(ValueError):
        sdl2ext.line(sf.contents, WHITE, (1, 2, 3))

def test_prepare_color(with_sdl):
    rcolors = (
        Color(0, 0, 0, 0),
        Color(255, 255, 255, 255),
        Color(8, 55, 110, 220),
    )
    icolors = (
        0x00000000,
        0xFFFFFFFF,
        0xAABBCCDD,
    )
    scolors = (
        "#000",
        "#FFF",
        "#AABBCCDD",
    )
    sf = _create_surface((10, 10), fmt="RGBA8888")
    for color in rcolors:
        c = sdl2ext.prepare_color(color, sf)
        assert c == int(color)
    for color in icolors:
        c = sdl2ext.prepare_color(color, sf)
        cc = COLOR(color)
        assert c == int(cc)
    for color in scolors:
        c = sdl2ext.prepare_color(color, sf)
        cc = COLOR(color)
        assert c == int(cc)
    SDL_FreeSurface(sf)


import sys
import warnings
try:
	from collections.abc import Callable, Iterable
except ImportError:
	from collections import Callable, Iterable

__all__ = ["ISPYTHON2", "ISPYTHON3", "utf8", "stringify", "byteify",
           "isiterable", "platform_is_64bit", "deprecated", "deprecation",
           "UnsupportedError", "ExperimentalWarning", "experimental",
           ]

ISPYTHON2 = False
ISPYTHON3 = False

if sys.version_info[0] < 3:
    ISPYTHON2 = True
else:
    __all__ += ["long", "unichr", "callable", "unicode"]
    ISPYTHON3 = True
    long = int
    unichr = chr
    callable = lambda x: isinstance(x, Callable)
    unicode = str


def _to_unicode(x, enc):
    if ISPYTHON2:
        if type(x) in (str, bytes):
            return x.decode(enc)
        else:
            return unicode(x)
    else:
        if type(x) == bytes:
            return x.decode(enc)
        else:
            return str(x)


def _is_text(x):
    return isinstance(x, (str, bytes, unicode))


def utf8(x):
    """Converts input to a unicode string in a Python 2/3 agnostic manner.

    If a :obj:`bytes` object is passed, it will be decoded as UTF-8. This
    function returns :obj:`unicode` for Python 2 and :obj:`str` for Python 3.

    Args:
        x: Input to convert to a unicode string.

    Returns:
        :obj:`str` on Python 3.x, or :obj:`unicode` on Python 2.7.

    """
    return _to_unicode(x, 'utf-8')


def stringify(x, enc='utf-8'):
    """Converts input to a :obj:`str` in a Python 2/3 agnostic manner.

    If the input is :obj:`unicode` and the Python version is 2.7, the ``enc``
    parameter indicates the encoding to use when converting the input to
    a non-unicode string. If the input is :obj:`bytes` and the Python version
    is 3.x, the ``enc`` parameter indicates the encoding to use to decode the
    input into a unicode string.
    
    Args:
        x: Input to convert to a :obj:`str`.
        enc (str, optional): The encoding type used to encode or decode the
            input, depending on the input type and the major Python version.
            Defaults to UTF-8.

    """
    if ISPYTHON2:
        if type(x) == unicode:
            return x.encode(enc)
        else:
            return str(x)
    else:
        return _to_unicode(x, enc)


def byteify(x, enc='utf-8'):
    """Converts input to :obj:`bytes` in a Python 2/3 agnostic manner.

    If the input is a unicode string, the ``enc`` parameter indicates
    the encoding to use when encoding the input to :obj:`bytes`.
    
    Args:
        x: Input to convert to :obj:`bytes`.
        enc (str, optional): The encoding type used to encode any unicode
            string input. Defaults to UTF-8.

    """
    unicode_str = unicode if ISPYTHON2 else str
    if type(x) == unicode_str:
        return x.encode(enc)
    else:
        return bytes(x)


def isiterable(x):
    """Checks whether the input is a non-string iterable.

    Args:
        x: The object to check for iterability.

    Returns:
        bool: True if the input is a valid iterable, otherwise False.

    """
    return hasattr(x, "__iter__") and not hasattr(x, "upper")


def platform_is_64bit():
    """Checks whether the Python interpreter is 64-bit.
    
    Returns:
        bool: True if running on 64-bit Python, otherwise False.

    """
    return sys.maxsize > 2 ** 32


def deprecated(func):
    # A simple decorator to mark functions and methods as deprecated
    def wrapper(*fargs, **kw):
        warnings.warn("%s is deprecated." % func.__name__,
                      category=DeprecationWarning, stacklevel=2)
        return func(*fargs, **kw)
    wrapper.__name__ = func.__name__
    wrapper.__doc__ = func.__doc__
    wrapper.__dict__.update(func.__dict__)
    return wrapper


def deprecation(message):
    # Prints a deprecation message
    warnings.warn(message, category=DeprecationWarning, stacklevel=2)


class UnsupportedError(RuntimeError):
    # Indicates that a certain class, function or behaviour is not supported
    pass


class ExperimentalWarning(Warning):
    # Indicates that a certain class, function or behaviour is experimental
    def __init__(self, obj, msg=None):
        """Creates a ExperimentalWarning for the specified obj.

        If a message is passed in msg, it will be printed instead of the
        default message.
        """
        super(ExperimentalWarning, self).__init__()
        self.obj = obj
        self.msg = msg

    def __str__(self):
        if self.msg is None:
            return "%s is in an experimental state." % repr(self.obj)
        return repr(self.msg)


def experimental(func):
    # A simple decorator to mark functions and methods as experimental
    def wrapper(*fargs, **kw):
        warnings.warn("%s" % func.__name__, category=ExperimentalWarning,
                      stacklevel=2)
        return func(*fargs, **kw)
    wrapper.__name__ = func.__name__
    wrapper.__doc__ = func.__doc__
    wrapper.__dict__.update(func.__dict__)
    return wrapper



from .. import surface, rect, video, pixels, render, rwops

from .color import convert_to_color
from .common import SDLError
from .compat import isiterable
from .ebs import System
from .image import load_image
from .renderer import Renderer
from .sprite import Sprite, SoftwareSprite, TextureSprite
from .window import Window

__all__ = [
    "SpriteFactory", "SoftwareSpriteRenderSystem", "SpriteRenderSystem",
    "TextureSpriteRenderSystem", "TEXTURE", "SOFTWARE"
]


TEXTURE = 0
SOFTWARE = 1


class SpriteFactory(object):
    """A factory class for creating Sprite components."""
    def __init__(self, sprite_type=TEXTURE, **kwargs):
        """Creates a new SpriteFactory.

        The SpriteFactory can create TextureSprite or SoftwareSprite
        instances, depending on the sprite_type being passed to it,
        which can be SOFTWARE or TEXTURE. The additional kwargs are used
        as default arguments for creating sprites within the factory
        methods.
        """
        if sprite_type == TEXTURE:
            if "renderer" not in kwargs:
                raise ValueError("you have to provide a renderer=<arg> argument")
        elif sprite_type != SOFTWARE:
            raise ValueError("sprite_type must be TEXTURE or SOFTWARE")
        self._spritetype = sprite_type
        self.default_args = kwargs

    @property
    def sprite_type(self):
        """The sprite type created by the factory."""
        return self._spritetype

    def __repr__(self):
        stype = "TEXTURE"
        if self.sprite_type == SOFTWARE:
            stype = "SOFTWARE"
        return "SpriteFactory(sprite_type=%s, default_args=%s)" % \
            (stype, self.default_args)

    def create_sprite_render_system(self, *args, **kwargs):
        """Creates a new SpriteRenderSystem.

        For TEXTURE mode, the passed args and kwargs are ignored and the
        Renderer or SDL_Renderer passed to the SpriteFactory is used.
        """
        if self.sprite_type == TEXTURE:
            return TextureSpriteRenderSystem(self.default_args["renderer"])
        else:
            return SoftwareSpriteRenderSystem(*args, **kwargs)

    def from_image(self, fname):
        """Creates a Sprite from the passed image file."""
        return self.from_surface(load_image(fname), True)

    def from_surface(self, tsurface, free=False):
        """Creates a Sprite from the passed SDL_Surface.

        If free is set to True, the passed surface will be freed
        automatically.
        """
        if self.sprite_type == TEXTURE:
            renderer = self.default_args["renderer"]
            texture = render.SDL_CreateTextureFromSurface(renderer.sdlrenderer,
                                                          tsurface)
            if not texture:
                raise SDLError()
            sprite = TextureSprite(texture.contents)
            if free:
                surface.SDL_FreeSurface(tsurface)
            return sprite
        elif self.sprite_type == SOFTWARE:
            return SoftwareSprite(tsurface, free)
        raise ValueError("sprite_type must be TEXTURE or SOFTWARE")

    def from_object(self, obj):
        """Creates a Sprite from an arbitrary object."""
        if self.sprite_type == TEXTURE:
            rw = rwops.rw_from_object(obj)
            # TODO: support arbitrary objects.
            imgsurface = surface.SDL_LoadBMP_RW(rw, True)
            if not imgsurface:
                raise SDLError()
            return self.from_surface(imgsurface.contents, True)
        elif self.sprite_type == SOFTWARE:
            rw = rwops.rw_from_object(obj)
            imgsurface = surface.SDL_LoadBMP_RW(rw, True)
            if not imgsurface:
                raise SDLError()
            return SoftwareSprite(imgsurface.contents, True)
        raise ValueError("sprite_type must be TEXTURE or SOFTWARE")

    def from_color(self, color, size, bpp=32, masks=None):
        """Creates a sprite with a certain color.
        """
        color = convert_to_color(color)
        if masks:
            rmask, gmask, bmask, amask = masks
        else:
            rmask = gmask = bmask = amask = 0
        sfc = surface.SDL_CreateRGBSurface(0, size[0], size[1], bpp, rmask,
                                           gmask, bmask, amask)
        if not sfc:
            raise SDLError()
        fmt = sfc.contents.format
        if fmt.contents.Amask != 0:
            # Target has an alpha mask
            col = pixels.SDL_MapRGBA(fmt, color.r, color.g, color.b, color.a)
        else:
            col = pixels.SDL_MapRGB(fmt, color.r, color.g, color.b)
        ret = surface.SDL_FillRect(sfc, None, col)
        if ret == -1:
            raise SDLError()
        return self.from_surface(sfc.contents, True)

    def from_text(self, text, **kwargs):
        """Creates a Sprite from a string of text."""
        args = self.default_args.copy()
        args.update(kwargs)
        fontmanager = args['fontmanager']
        sfc = fontmanager.render(text, **args)
        return self.from_surface(sfc, free=True)

    def create_sprite(self, **kwargs):
        """Creates an empty Sprite.

        This will invoke create_software_sprite() or
        create_texture_sprite() with the passed arguments and the set
        default arguments.
        """
        args = self.default_args.copy()
        args.update(kwargs)
        if self.sprite_type == TEXTURE:
            return self.create_texture_sprite(**args)
        else:
            return self.create_software_sprite(**args)

    def create_software_sprite(self, size, bpp=32, masks=None):
        """Creates a software sprite.

        A size tuple containing the width and height of the sprite and a
        bpp value, indicating the bits per pixel to be used, need to be
        provided.
        """
        if masks:
            rmask, gmask, bmask, amask = masks
        else:
            rmask = gmask = bmask = amask = 0
        imgsurface = surface.SDL_CreateRGBSurface(0, size[0], size[1], bpp,
                                                  rmask, gmask, bmask, amask)
        if not imgsurface:
            raise SDLError()
        return SoftwareSprite(imgsurface.contents, True)

    def create_texture_sprite(self, renderer, size,
                              pformat=pixels.SDL_PIXELFORMAT_RGBA8888,
                              access=render.SDL_TEXTUREACCESS_STATIC):
        """Creates a texture sprite.

        A size tuple containing the width and height of the sprite needs
        to be provided.

        TextureSprite objects are assumed to be static by default,
        making it impossible to access their pixel buffer in favour for
        faster copy operations. If you need to update the pixel data
        frequently or want to use the texture as target for rendering
        operations, access can be set to the relevant
        SDL_TEXTUREACCESS_* flag.
        """
        if isinstance(renderer, render.SDL_Renderer):
            sdlrenderer = renderer
        elif isinstance(renderer, Renderer):
            sdlrenderer = renderer.sdlrenderer
        else:
            raise TypeError("renderer must be a Renderer or SDL_Renderer")
        texture = render.SDL_CreateTexture(sdlrenderer, pformat, access,
                                           size[0], size[1])
        if not texture:
            raise SDLError()
        return TextureSprite(texture.contents)


class SpriteRenderSystem(System):
    """A rendering system for Sprite components.

    This is a base class for rendering systems capable of drawing and
    displaying Sprite-based objects. Inheriting classes need to
    implement the rendering capability by overriding the render()
    method.
    """
    def __init__(self):
        super(SpriteRenderSystem, self).__init__()
        self.componenttypes = (Sprite,)
        self._sortfunc = lambda e: e.depth

    def render(self, sprites, x=None, y=None):
        """Renders the passed sprites.

        This is a no-op function and needs to be implemented by inheriting
        classes.
        """
        pass

    def process(self, world, components):
        """Draws the passed SoftSprite objects on the Window's surface."""
        self.render(sorted(components, key=self._sortfunc))

    @property
    def sortfunc(self):
        """Sort function for the component processing order.

        The default sort order is based on the depth attribute of every
        sprite. Lower depth values will cause sprites to be drawn below
        sprites with higher depth values.
        """
        return self._sortfunc

    @sortfunc.setter
    def sortfunc(self, value):
        """Sort function for the component processing order.

        The default sort order is based on the depth attribute of every
        sprite. Lower depth values will cause sprites to be drawn below
        sprites with higher depth values.
        """
        if not callable(value):
            raise TypeError("sortfunc must be callable")
        self._sortfunc = value


class SoftwareSpriteRenderSystem(SpriteRenderSystem):
    """A rendering system for SoftwareSprite components.

    The SoftwareSpriteRenderSystem class uses a Window as drawing device to
    display Sprite surfaces. It uses the Window's internal SDL surface as
    drawing context, so that GL operations, such as texture handling or
    using SDL renderers is not possible.
    """
    def __init__(self, window):
        """Creates a new SoftwareSpriteRenderSystem for a specific Window."""
        super(SoftwareSpriteRenderSystem, self).__init__()
        if isinstance(window, Window):
            self.window = window.window
        elif isinstance(window, video.SDL_Window):
            self.window = window
        else:
            raise TypeError("unsupported window type")
        self.target = window
        sfc = video.SDL_GetWindowSurface(self.window)
        if not sfc:
            raise SDLError()
        self.surface = sfc.contents
        self.componenttypes = (SoftwareSprite,)

    def render(self, sprites, x=None, y=None):
        """Draws the passed sprites (or sprite) on the Window's surface.

        x and y are optional arguments that can be used as relative drawing
        location for sprites. If set to None, the location information of the
        sprites are used. If set and sprites is an iterable, such as a list of
        SoftwareSprite objects, x and y are relative location values that will
        be added to each individual sprite's position. If sprites is a single
        SoftwareSprite, x and y denote the absolute position of the
        SoftwareSprite, if set.
        """
        r = rect.SDL_Rect(0, 0, 0, 0)
        if isiterable(sprites):
            blit_surface = surface.SDL_BlitSurface
            imgsurface = self.surface
            x = x or 0
            y = y or 0
            for sprite in sprites:
                r.x = x + sprite.x
                r.y = y + sprite.y
                blit_surface(sprite.surface, None, imgsurface, r)
        else:
            r.x = sprites.x
            r.y = sprites.y
            if x is not None and y is not None:
                r.x = x
                r.y = y
            surface.SDL_BlitSurface(sprites.surface, None, self.surface, r)
        video.SDL_UpdateWindowSurface(self.window)


class TextureSpriteRenderSystem(SpriteRenderSystem):
    """A rendering system for TextureSprite components.

    The TextureSpriteRenderSystem class uses a SDL_Renderer as drawing
    device to display TextureSprite objects.
    """
    def __init__(self, target):
        """Creates a new TextureSpriteRenderSystem.

        target can be a Window, SDL_Window, Renderer or SDL_Renderer.
        If it is a Window or SDL_Window instance, a Renderer will be
        created to acquire the SDL_Renderer.
        """
        super(TextureSpriteRenderSystem, self).__init__()
        if isinstance(target, (Window, video.SDL_Window)):
            # Create a Renderer for the window and use that one.
            target = Renderer(target)

        if isinstance(target, Renderer):
            self._renderer = target  # Used to prevent GC
            sdlrenderer = target.sdlrenderer
        elif isinstance(target, render.SDL_Renderer):
            sdlrenderer = target
        else:
            raise TypeError("unsupported object type")
        self.sdlrenderer = sdlrenderer
        self.componenttypes = (TextureSprite,)

    def __del__(self):
        self.sdlrenderer = None
        if hasattr(self, "_renderer"):
            self._renderer = None

    def render(self, sprites, x=None, y=None):
        """Draws the passed sprites (or sprite).

        x and y are optional arguments that can be used as relative
        drawing location for sprites. If set to None, the location
        information of the sprites are used. If set and sprites is an
        iterable, such as a list of TextureSprite objects, x and y are
        relative location values that will be added to each individual
        sprite's position. If sprites is a single TextureSprite, x and y
        denote the absolute position of the TextureSprite, if set.
        """
        r = rect.SDL_Rect(0, 0, 0, 0)
        rcopy = render.SDL_RenderCopyEx
        if isiterable(sprites):
            renderer = self.sdlrenderer
            x = x or 0
            y = y or 0
            for sprite in sprites:
                r.x = x + sprite.x
                r.y = y + sprite.y
                r.w, r.h = sprite.size
                if rcopy(renderer, sprite.texture, None, r, sprite.angle,
                         sprite.center, sprite.flip) == -1:
                    raise SDLError()
        else:
            r.x = sprites.x
            r.y = sprites.y
            r.w, r.h = sprites.size
            if x is not None and y is not None:
                r.x = x
                r.y = y
            if rcopy(self.sdlrenderer, sprites.texture, None, r, sprites.angle,
                     sprites.center, sprites.flip) == -1:
                raise SDLError()
        render.SDL_RenderPresent(self.sdlrenderer)



import sys
import pytest
from ctypes import c_float 
import sdl2
from sdl2 import SDL_Init, SDL_Quit, SDL_INIT_SENSOR
from sdl2.events import SDL_QUERY, SDL_ENABLE, SDL_IGNORE
from sdl2.stdinc import SDL_TRUE, SDL_FALSE
from sdl2.error import SDL_GetError, SDL_ClearError
from sdl2 import sensor

# Make sure sensor subsystem is available and works before running tests
available = sdl2.dll.version >= 2009
ret = SDL_Init(SDL_INIT_SENSOR)
SDL_Quit()
skipmsg = 'Sensor subsystem not supported'
pytestmark = pytest.mark.skipif(ret != 0 or not available, reason=skipmsg)


def test_SDL_NumSensors():
    ret = SDL_Init(SDL_INIT_SENSOR)
    assert ret == 0
    retval = sensor.SDL_NumSensors()
    SDL_Quit()
    assert retval >= 0


skipmsg = "sensor subsystem not supported"
@pytest.mark.skipif(SDL_Init(SDL_INIT_SENSOR) != 0, reason=skipmsg)
class TestSDLSensor(object):
    __tags__ = ["sdl"]

    @classmethod
    def setup_class(cls):
        SDL_Init(SDL_INIT_SENSOR)
        num = sensor.SDL_NumSensors()
        if num < 1:
            pytest.skip("no available sensor devices")
        cls.scount = num

    @classmethod
    def teardown_class(cls):
        SDL_Quit()

    def setup_method(self):
        SDL_ClearError()

    def test_SDL_SensorOpenClose(self):
        for index in range(self.scount):
            s = sensor.SDL_SensorOpen(index)
            assert isinstance(s.contents, sensor.SDL_Sensor)
            sensor.SDL_SensorClose(s)

    def test_SDL_SensorFromInstanceID(self):
        for index in range(self.scount):
            iid = sensor.SDL_SensorGetDeviceInstanceID(index)
            s = sensor.SDL_SensorFromInstanceID(iid)
            assert isinstance(s.contents, sensor.SDL_Sensor)
            sensor.SDL_SensorClose(s)

    def test_SDL_SensorUpdate(self):
        sensor.SDL_SensorUpdate()

    def test_SDL_SensorDeviceInfo(self):
        types = {
            sensor.SDL_SENSOR_INVALID: "Invalid",
            sensor.SDL_SENSOR_UNKNOWN: "Unknown",
            sensor.SDL_SENSOR_ACCEL: "Accelerometer",
            sensor.SDL_SENSOR_GYRO: "Gyroscope"
        }
        for index in range(self.scount):
            name = sensor.SDL_SensorGetDeviceName(index)
            assert type(name) in (str, bytes)
            dtype = sensor.SDL_SensorGetDeviceType(index)
            assert dtype in list(types.keys())
            nptype = sensor.SDL_SensorGetDeviceNonPortableType(index)
            assert type(nptype) == int
            iid = sensor.SDL_SensorGetDeviceInstanceID(index)
            assert type(iid) == int

    def test_SDL_SensorInfo(self):
        sensors = []
        types = {
            sensor.SDL_SENSOR_INVALID: "Invalid",
            sensor.SDL_SENSOR_UNKNOWN: "Unknown",
            sensor.SDL_SENSOR_ACCEL: "Accelerometer",
            sensor.SDL_SENSOR_GYRO: "Gyroscope"
        }
        for index in range(self.scount):
            dtype = sensor.SDL_SensorGetDeviceType(index)
            if dtype in [sensor.SDL_SENSOR_INVALID, sensor.SDL_SENSOR_UNKNOWN]:
                continue
            s = sensor.SDL_SensorOpen(index)
            name = sensor.SDL_SensorGetName(s)
            assert type(name) in (str, bytes)
            dtype = sensor.SDL_SensorGetType(s)
            assert dtype in list(types.keys())
            nptype = sensor.SDL_SensorGetNonPortableType(s)
            assert type(nptype) == int
            iid = sensor.SDL_SensorGetInstanceID(s)
            assert type(iid) == int
            sensor.SDL_SensorUpdate()
            data = c_float(-0.00323)  # Set a random initial value
            assert sensor.SDL_SensorGetData(s, data, 1) == 0
            assert data.value != -0.00323
            sensor.SDL_SensorClose(s)
            s_info = {
                'name': name.decode('utf-8'), 'type': types[dtype],
                'np': nptype, 'id': iid, 'data': data.value
            }
            sensors.append(s_info)
        if len(sensors):
            print("Available sensors:")
        for s in sensors:
            print("  '{0}' (ID: {1}):".format(s['name'], s['id']))
            print("   - Sensor Type: {0}".format(s['type']))
            print("   - Non-Portable Type: {0}".format(s['nptype']))
            print("   - Current X Value: {0}".format(s['data']))
            
    @pytest.mark.skipif(sdl2.dll.version < 2014, reason="not available")
    def test_SDL_LockUnlockSensors(self):
        # NOTE: not sure how better to test these, since I don't know if
        # they'd even be useful at all in Python given the GIL
        sensor.SDL_LockSensors()
        sensor.SDL_UnlockSensors()



import sys
from ctypes import c_int, byref
import pytest
import sdl2
from sdl2 import SDL_Init, SDL_Quit, SDL_INIT_VIDEO
from sdl2 import metal, video, error

macos = sys.platform == "darwin"


# TODO: Add more complete tests with pyobjc

@pytest.mark.xfail(reason="Metal not supported on all macs")
@pytest.mark.skipif(sdl2.dll.version < 2012 or not macos, reason="not available")
def test_SDL_Metal_CreateDestroyView(with_sdl):
    flags = video.SDL_WINDOW_HIDDEN | video.SDL_WINDOW_METAL
    window = video.SDL_CreateWindow(b"Test", 10, 10, 10, 10, flags)
    assert isinstance(window.contents, video.SDL_Window)
    view = metal.SDL_Metal_CreateView(window)
    err = error.SDL_GetError()
    if len(err):
        print("Metal Error: '{0}'".format(err.decode('utf-8')))
    assert view  # Verify pointer is not null
    metal.SDL_Metal_DestroyView(view)
    video.SDL_DestroyWindow(window)

@pytest.mark.xfail(reason="Metal not supported on all macs")
@pytest.mark.skipif(sdl2.dll.version < 2014 or not macos, reason="not available")
def test_SDL_Metal_GetLayer(with_sdl):
    flags = video.SDL_WINDOW_HIDDEN | video.SDL_WINDOW_METAL
    window = video.SDL_CreateWindow(b"Test", 10, 10, 10, 10, flags)
    assert isinstance(window.contents, video.SDL_Window)
    view = metal.SDL_Metal_CreateView(window)
    assert view  # Verify pointer is not null
    layer = metal.SDL_Metal_GetLayer(view)
    assert layer  # Verify pointer is not null
    metal.SDL_Metal_DestroyView(view)
    video.SDL_DestroyWindow(window)

@pytest.mark.xfail(reason="Metal not supported on all macs")
@pytest.mark.skipif(sdl2.dll.version < 2014 or not macos, reason="not available")
def test_SDL_Metal_GetDrawableSize(with_sdl):
    flags = video.SDL_WINDOW_HIDDEN | video.SDL_WINDOW_METAL
    window = video.SDL_CreateWindow(b"Test", 10, 10, 10, 10, flags)
    assert isinstance(window.contents, video.SDL_Window)
    view = metal.SDL_Metal_CreateView(window)
    assert view  # Verify pointer is not null
    w, h = c_int(0), c_int(0)
    metal.SDL_Metal_GetDrawableSize(window, byref(w), byref(h))
    assert w.value == 10 and h.value == 10
    metal.SDL_Metal_DestroyView(view)
    video.SDL_DestroyWindow(window)
